import json
import sys
import os
import bz2
from itertools import chain
from contextlib import contextmanager

try:
    from tqdm import tqdm
    TQDM_IMPORT_ERROR = None
except ImportError as err:
    tqdm = None
    TQDM_IMPORT_ERROR = err

from .. import MarkovBase, MarkovJsonMixin, MarkovSqliteMixin


JSON = 0
SQLITE = 1
IJSON_MIN_SIZE = 4 * 1024 * 1024
IJSON_MIN_COMPRESSED_SIZE = 512 * 1024

BAR_DESC_SIZE = 12
BAR_N_SIZE = 8
BAR_RATE_SIZE = 14
BAR_FORMAT = '{{desc:<{0}.{0}}}{{percentage:3.0f}}%' \
             '|{{bar}}| ' \
             '{{n_fmt:>{1}.{1}}}/{{total_fmt:<{1}.{1}}} '\
             '{{elapsed}}<{{remaining:^5}} {{rate_fmt:>{2}.{2}}}' \
                 .format(BAR_DESC_SIZE, BAR_N_SIZE, BAR_RATE_SIZE)


class NoProgressBar:
    """Missing progress bar class.

    Attributes
    ----------
    warning : `bool`
        True if a missing progress bar warning was printed.
    """
    warning = False

    @classmethod
    def print_warning(cls):
        """Print a missing progress bar warning if it was not printed.
        """
        if not cls.warning:
            cls.warning = True
            print('Can\'t create progress bar:', str(TQDM_IMPORT_ERROR),
                  file=sys.stderr)

    def update(self, *args, **kwargs):
        """Do nothing.
        """
        pass

    def close(self, *args, **kwargs):
        """Do nothing.
        """
        pass


def no_tqdm(iterable=None, *args, **kwargs): # pylint: disable=unused-argument
    """Print a missing progress bar warning if it was not printed.

    Parameters
    ----------
    iterable : `iterable` or `None`, optional
        Iterable to decorate with a progress bar (default: None).

    Returns
    -------
    `iterable` or `markovchain.cli.util.NoProgressBar`
    """
    NoProgressBar.print_warning()
    if iterable is not None:
        return iterable
    return NoProgressBar()

if tqdm is None:
    tqdm = no_tqdm # pylint: disable=invalid-name

def pprint(data, indent=0, end='\n'):
    """Pretty print JSON data.

    Parameters
    ----------
    data
        JSON data.
    indent : `int`, optional
        Indent level in characters (default: 0).
    end : `str`, optional
        String to print after the data (default: '\\\\n').
    """
    if isinstance(data, dict):
        print('{')
        new_indent = indent + 4
        space = ' ' * new_indent
        keys = list(sorted(data.keys()))
        for i, k in enumerate(keys):
            print(space, json.dumps(k), ': ', sep='', end='')
            pprint(data[k], new_indent,
                   end=',\n' if i < len(keys) - 1 else '\n')
        print(' ' * indent, '}', sep='', end=end)
    elif isinstance(data, list):
        if any(isinstance(x, (dict, list)) for x in data):
            print('[')
            new_indent = indent + 4
            space = ' ' * new_indent
            for i, x in enumerate(data):
                print(space, end='')
                pprint(x, new_indent,
                       end=',\n' if i < len(data) - 1 else '\n')
            print(' ' * indent, ']', sep='', end=end)
        else:
            print(json.dumps(data), end=end)
    else:
        print(json.dumps(data), end=end)

def load(markov, fname, args):
    """Load a markov chain generator.

    Parameters
    ----------
    markov : `type`
        Markov chain generator class.
    fname : `str`
        Input file path.
    args : `argparse.Namespace`
        Command arguments.
    """
    if issubclass(markov, MarkovJsonMixin):
        size = os.path.getsize(fname)

        if fname.endswith('bz2'):
            op = bz2.open
            maxsize = IJSON_MIN_COMPRESSED_SIZE
        else:
            op = open
            maxsize = IJSON_MIN_SIZE

        if size > maxsize:
            mode = 'rb'
        else:
            mode = 'rt'

        if args.progress:
            print('Loading JSON data...')

        with op(fname, mode) as fp:
            return markov.load(fp, args.settings)
    else:
        return markov.load(fname, args.settings)

def save(markov, fname, args):
    """Save a markov chain generator.

    Parameters
    ----------
    markov : `MarkovBase`
        Markov chain generator.
    fname : `str`
        Output file path.
    args : `argparse.Namespace`
        Command arguments.
    """
    if isinstance(markov, MarkovJsonMixin):
        if fname is None:
            markov.save(sys.stdout)
        else:
            if fname.endswith('bz2'):
                op = bz2.open
            else:
                op = open

            if args.progress:
                print('Saving JSON data...')

            with op(fname, 'wt') as fp:
                markov.save(fp)
    else:
        markov.save()

def set_args(args, base):
    """Set computed command arguments.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    base : `iterable` of `type`
        Generator mixins.

    Raises
    ------
    ValueError
        If output file is stdout and progress bars are enabled.
    """
    try:
        if args.output is sys.stdout and args.progress:
            raise ValueError('args.output is stdout and args.progress')
    except AttributeError:
        pass

    try:
        fname = '.' + args.type
    except AttributeError:
        try:
            fname = args.state
        except AttributeError:
            try:
                fname = args.output
            except AttributeError:
                fname = '.json'

    if fname is None or fname.endswith('.json') or fname.endswith('.bz2'):
        args.type = JSON
    else:
        args.type = SQLITE

    if args.type == JSON:
        dtype = MarkovJsonMixin
    elif args.type == SQLITE:
        dtype = MarkovSqliteMixin

    base = tuple(chain((dtype,), base, (MarkovBase,)))

    class Markov(*base): # pylint: disable=too-few-public-methods
        pass

    try:
        settings = args.settings
        has_settings = True
    except AttributeError:
        settings = None
        has_settings = False

    if has_settings:
        if settings is not None:
            settings = json.load(settings)
            args.settings.close()
        else:
            settings = {}

    args.settings = settings
    args.markov = Markov

def check_output_format(fmt, nfiles):
    """Validate file format string.

    Parameters
    ----------
    fmt : `str`
        File format string.
    nfiles : `int`
        Number of files.

    Raises
    ------
    ValueError
        If nfiles < 0 or format string is invalid.
    """
    if nfiles < 0:
        raise ValueError('Invalid file count: ' + str(nfiles))
    if nfiles == 1:
        return
    try:
        fmt % nfiles
    except TypeError as err:
        raise ValueError(''.join(
            ('Invalid file format string: ', fmt, ': ', str(err))
        ))

@contextmanager
def infiles(fnames, progress, leave=True):
    """Get input file paths.

    Parameters
    ----------
    fnames : `list` of `str`
        File paths.
    progress : `bool`
        Show progress bar.
    leave : `bool`, optional
        Leave progress bar (default: True).

    Returns
    -------
    `generator` of `str`
        Input file paths.
    """
    if progress:
        if fnames:
            fnames = tqdm(fnames, desc='Loading', unit='file',
                          bar_format=BAR_FORMAT,
                          leave=leave, dynamic_ncols=True)
        else:
            progress = False

    yield fnames

    if progress:
        fnames.close()

@contextmanager
def outfiles(fmt, nfiles, progress, leave=True):
    """Get output file paths.

    Parameters
    ----------
    fmt : `str`
        File path format string.
    nfiles : `int`
        Number of files.
    progress : `bool`
        Show progress bars.
    leave : `bool`, optional
        Leave progress bar (default: True).

    Raises
    ------
    ValueError
        If nfiles <= 0.

    Returns
    -------
    `generator` of `str`
        Output file paths.
    """
    if nfiles > 1:
        fnames = (fmt % i for i in range(nfiles))
    elif nfiles == 1:
        fnames = (fmt,)
    else:
        raise ValueError('output file count <= 0')

    if progress:
        fnames = tqdm(fnames, total=nfiles,
                      desc='Generating', unit='file',
                      bar_format=BAR_FORMAT,
                      leave=leave, dynamic_ncols=True)

    yield fnames

    if progress:
        fnames.close()

def cmd_settings(args):
    """Print generator settings.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    markov = load(args.markov, args.state, args)
    data = markov.get_save_data()
    try:
        del data['nodes']
    except KeyError:
        pass
    pprint(data)



import re

from .util import SaveLoad


class Scanner(SaveLoad):
    """Base scanner class.

    Attributes
    ----------
    classes : `dict`
        Scanner class group.
    START
        Sentence start token.
    END
        Sentence end token.

    Examples
    --------
    >>> scan = Scanner(lambda data: data.split())
    >>> scan('a b c')
    ['a', 'b', 'c']
    """

    classes = {}

    END = None
    START = None

    def __init__(self, scan=None):
        """Base scanner constructor.

        Parameters
        ----------
        scan : `function`, optional
        """
        if scan is not None:
            self.scan = scan

    def __call__(self, data, part=False):
        """Scan data.

        Parameters
        ----------
        data
            Data to scan.
        part : `bool`, optional
            `True` if data is partial (default: `False`).

        Returns
        -------
        `object`
            self.scan(data)
        """
        return self.scan(data)

    def reset(self):
        """Reset scanner state.
        """
        pass


class CharScanner(Scanner):
    """Character scanner.

    Attributes
    ----------
    end_chars : `str`
        Sentence ending characters.
    default_end : `str`
        Default sentence ending character.
    start : `bool`
        True if current sentence is started.
    end : `bool`
        True if current sentence is ended.

    Examples
    --------
    >>> scan = CharScanner()
    >>> list(scan('Word'))
    ['W', 'o', 'r', 'd', '.', Scanner.END]
    >>> list(scan('Word', True))
    ['W', 'o', 'r', 'd']
    >>> list(scan(''))
    ['.', Scanner.END]
    """
    def __init__(self, end_chars='.?!', default_end='.'):
        """Character scanner constructor.

        Parameters
        ----------
        end_chars : `str`, optional
            Sentence ending characters (default: '.?!').
        default_end : `str`, optional
            Default sentence ending character (default: '.').
        """
        super().__init__()
        self.end_chars = end_chars
        self.default_end = default_end
        self.start = False
        self.end = False

    def reset(self):
        """Reset scanner state.
        """
        self.start = False
        self.end = False

    def __call__(self, data, part=False):
        """Scan a string.

        Parameters
        ----------
        data : `str`
            String to scan.
        part : `bool`, optional
            True if data is partial (default: `False`).

        Returns
        -------
        `generator` of (`str` or `markovchain.scanner.Scanner.END`)
            Token generator.
        """
        if not self.end_chars:
            yield from data
            self.start = self.start or bool(data)
            self.end = False
        else:
            for char in data:
                if char in self.end_chars:
                    if not self.start:
                        continue
                    self.end = True
                else:
                    if self.end:
                        yield self.END
                        self.end = False
                    self.start = True
                yield char

        if not part and self.start:
            if not self.end and self.default_end is not None:
                yield self.default_end
            yield self.END
            self.reset()

    def __eq__(self, scanner):
        return (self.end_chars == scanner.end_chars
                and self.default_end == scanner.default_end)

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['end_chars'] = self.end_chars
        data['default_end'] = self.default_end
        return data


class RegExpScanner(Scanner):
    """Regular expression scanner.

    Attributes
    ----------
    DEFAULT_EXPR : `re._sre.SRE_Pattern`
        Default regular expression.
    expr : `re._sre.SRE_Pattern`
        Regular expression..
    default_end : `str`
        Default sentence ending string.
    end : `bool`
        `True` if current sentence is ended.

    Examples
    --------
    >>> scan = RegExpScanner(lambda data: data.split())
    >>> list(scan('Word word. word'))
    ['Word', 'word', '.', Scanner.END, 'word', '.', Scanner.END]
    >>> list(scan('word', True))
    ['word']
    >>> list(scan(''))
    ['.', Scanner.END]
    """

    DEFAULT_EXPR = re.compile(
        r'(?:(?P<end>[.!?]+)|(?P<word>(?:[^\w\s]+|\w+)))'
    )

    def __init__(self, expr=DEFAULT_EXPR, default_end='.'):
        """Regular expression scanner constructor.

        Parameters
        ----------
        expr : `str` or `re._sre.SRE_Pattern`, optional
            Regular expression (default: `markovchain.scanner.RegExpScanner.DEFAULT_EXPR`).
            It should have groups 'end' (sentence ending punctuation)
            and 'word' (words / other punctuation).
        default_end : `str`, optional
            Default sentence ending string (default: '.').
        """
        super().__init__()
        self.expr = self.get_regexp(expr)
        self.default_end = default_end
        self.end = True

    def reset(self):
        """Reset scanner state.
        """
        self.end = True

    def __call__(self, data, part=False):
        """Scan a string.

        Parameters
        ----------
        data : `str`
            String to scan.
        part : `bool`, optional
            `True` if data is partial (default: `False`).

        Returns
        -------
        `generator` of (`str` or `markovchain.scanner.Scanner.END`)
            Token generator.
        """
        if not self.expr.groups:
            for match in self.expr.finditer(data):
                yield match.group()

            self.end = self.end and not bool(data)
        else:
            for match in self.expr.finditer(data):
                group = self.get_group(match, 'end')
                if group is not None:
                    if not self.end:
                        yield group
                        yield self.END
                        self.end = True
                else:
                    self.end = False
                    group = self.get_group(match, 'word')
                    if group is not None:
                        yield group
                    else:
                        yield match.group()

        if not part and not self.end:
            if self.default_end is not None:
                yield self.default_end
            yield self.END
            self.reset()

    def __eq__(self, scanner):
        return (self.expr == scanner.expr
                and self.default_end == scanner.default_end)

    def save(self):
        """Convert the scanner to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['expr'] = self.expr.pattern
        data['default_end'] = self.default_end
        return data

    @staticmethod
    def get_regexp(x):
        """Compile a regular expression if necessary.

        Parameters
        ----------
        x : `str` or `re._sre.SRE_Pattern`
            Regular expression.

        Returns
        -------
        `re._sre.SRE_Pattern`
            Compiled regular expression.
        """
        if isinstance(x, str):
            return re.compile(x)
        return x

    @staticmethod
    def get_group(match, group):
        """Get a group from a regular expression match object if it exists.

        Parameters
        ----------
        match : `re._sre.SRE_Match`
            Regular expression match object.
        group : `str` or `int`
            Group name or index.

        Returns
        -------
        `str` or `None`
        """
        try:
            return match.group(group)
        except IndexError:
            return None



from unittest import TestCase
from PIL import Image

from markovchain import Scanner
from markovchain.image import ImageScanner, HLines, VLines
from markovchain.image.util import palette as default_palette


class TestImageScanner(TestCase):
    @classmethod
    def setUpClass(cls):
        data = b'\x00\x00\x00\xaa\xaa\xaa\xdd\xdd\xdd\x44\x44\x44'
        palette = [
            0x00, 0x00, 0x00,
            0x44, 0x44, 0x44,
            0xaa, 0xaa, 0xaa,
            0xdd, 0xdd, 0xdd
        ]
        palette.extend(0 for _ in range((256 - len(palette)) * 3))
        cls.image = Image.frombytes('RGB', (2, 2), data, 'raw', 'RGB')
        cls.palette = palette

    def test_properties(self):
        scan = ImageScanner()

        self.assertEqual(scan.palette, default_palette(8, 4, 8))
        self.assertIsNotNone(scan.palette_image)
        self.assertIsNotNone(scan.levels)
        self.assertIsNotNone(scan.min_size)
        self.assertIsInstance(scan.level_scale, list)
        self.assertIsInstance(scan.traversal, list)
        self.assertEqual(len(scan.level_scale), scan.levels - 1)
        self.assertEqual(len(scan.traversal), scan.levels)

        scan.palette = self.palette
        self.assertIsNotNone(scan.palette)
        self.assertIsNotNone(scan.palette_image)

        with self.assertRaises(ValueError):
            scan.levels = 0
        scan.levels = 3
        self.assertEqual(len(scan.traversal), scan.levels)

        with self.assertRaises(ValueError):
            scan.level_scale = []
        with self.assertRaises(ValueError):
            scan.level_scale = [1, -1]
        scan.level_scale = range(2, 10)
        self.assertEqual(scan.level_scale, [2, 3])
        self.assertEqual(scan.min_size, 6)
        scan.level_scale = 2
        self.assertEqual(scan.level_scale, [2, 2])
        self.assertEqual(scan.min_size, 4)
        scan.levels = 2
        self.assertEqual(len(scan.traversal), scan.levels)
        self.assertEqual(scan.level_scale, [2])
        self.assertEqual(scan.min_size, 2)
        scan.levels = 1
        self.assertEqual(len(scan.traversal), scan.levels)
        self.assertEqual(scan.level_scale, [])
        self.assertEqual(scan.min_size, 1)

        traversal = [HLines(), VLines()]
        scan.traversal = traversal
        scan.levels = 2
        self.assertIs(scan.traversal, traversal)
        self.assertEqual(scan.level_scale, [2])

    def test_input(self):
        tests = [(1, 1), (4, 1), (1, 4)]
        for test in tests:
            scan = ImageScanner(palette=self.palette, resize=test)
            img = scan.input(self.image)
            self.assertLessEqual(img.size, test)
            self.assertEqual(img.mode, 'RGB')
            img = [scan.level(img, level) for level in range(scan.levels)]
            self.assertEqual(len(img), 1)
            img = img[0]
            self.assertEqual(img.mode, 'P')
            self.assertEqual(img.size, (1, 1))
            self.assertIn(list(img.getdata()), [[0], [1]])

    def test_input_levels(self):
        scan = ImageScanner(palette=self.palette, levels=2, level_scale=2)

        img = scan.input(self.image)
        img = [scan.level(img, level) for level in range(scan.levels)]

        self.assertEqual(len(img), 2)

        self.assertEqual(img[0].mode, 'P')
        self.assertEqual(img[0].size, (1, 1))
        self.assertIn(list(img[0].getdata()), [[0], [1]])

        self.assertEqual(img[1].mode, 'P')
        self.assertEqual(img[1].size, (2, 2))
        self.assertEqual(list(img[1].getdata()), [0, 2, 3, 1])

        scan = ImageScanner(palette=self.palette, levels=3, level_scale=2)
        with self.assertRaises(ValueError):
            scan.input(self.image)

    def test_input_level_scale(self):
        img = Image.new(mode='RGB', size=(48, 48))
        scan = ImageScanner(palette=self.palette,
                            levels=4, level_scale=[2, 3, 4])

        img = scan.input(img)
        size = [scan.level(img, level).size for level in range(scan.levels)]

        self.assertEqual(size, [(2, 2), (4, 4), (12, 12), (48, 48)])

    def test_scan(self):
        scan = ImageScanner(palette=self.palette, traversal=HLines())
        self.assertEqual([list(level) for level in scan(self.image)],
                         [['00', '02', '03', '01', scan.END]])

    def test_scan_levels(self):
        scan = ImageScanner(palette=self.palette,
                            levels=2, level_scale=2,
                            traversal=[HLines(), VLines()])
        self.assertIn(
            [list(level) for level in scan(self.image)],
            [
                [
                    ['00', scan.END],
                    [(scan.START, '0000'),
                     '0100', '0103', '0102', '0101',
                     scan.END]
                ],
                [
                    ['01', scan.END],
                    [(scan.START, '0001'),
                     '0100', '0103', '0102', '0101',
                     scan.END]
                ],
            ]
        )

    def test_save_load(self):
        tests = [
            (),
            ((4, 4), 0, True, self.palette, 2, 2,
             Image.NEAREST, [HLines(), VLines()])
        ]
        for test in tests:
            scanner = ImageScanner(*test)
            saved = scanner.save()
            loaded = Scanner.load(saved)
            self.assertEqual(scanner, loaded)



import json
from random import randint

try:
    import ijson.backends.yajl2_cffi as ijson
except ImportError:
    try:
        import ijson.backends.yajl2 as ijson
    except ImportError:
        try:
            import ijson
        except ImportError:
            ijson = None

from .util import extend


class MarkovJsonMixin:
    """Markov chain JSON data mixin.

    Attributes
    ----------
    nodes : `dict` of ((`str`, `int`) or (`list` of `str`, `list` of `int`))
    """
    def __init__(self, nodes=None, *args, **kwargs):
        """Markov chain JSON data constructor.

        Attributes
        ----------
            nodes : `dict` of ((`str`, `int`) or (`list` of `str`, `list` of `int`)), optional
        """
        super().__init__(*args, **kwargs)
        if not nodes:
            self.nodes = {}
        else:
            self.nodes = nodes

    def __eq__(self, markov):
        return (self.nodes == markov.nodes
                and super().__eq__(markov))

    def replace_state_separator(self, old_separator, new_separator):
        """Replace state separator.

        Parameters
        ----------
        old_separator : `str`
            Old state separator.
        new_separator : `str`
            New state separator.
        """
        self.nodes = dict(
            (k.replace(old_separator, new_separator), v)
            for k, v in self.nodes.items()
        )
        #for node, (nodes, values) in self.nodes.items():
        #    if isinstance(nodes, list):
        #        for i, k in enumerate(nodes):
        #            nodes[i] = k.replace(old_separator, new_separator)
        #    else:
        #        self.nodes[node] = (
        #            nodes.replace(old_separator, new_separator),
        #            values
        #        )

    def links(self, links):
        """Update links.

        Parameters
        ----------
        links : `generator` of (`islice` of `str`, `str`)
            Links to add.
        """
        for src, dst in links:
            src = self.separator.join(src)
            try:
                node = self.nodes[src]
                nodes, values = node
                if isinstance(nodes, list):
                    try:
                        idx = nodes.index(dst)
                        values[idx] += 1
                    except ValueError:
                        nodes.append(dst)
                        values.append(1)
                elif nodes == dst:
                    node[1] += 1
                else:
                    node[0] = [nodes, dst]
                    node[1] = [values, 1]
            except KeyError:
                self.nodes[src] = [dst, 1]

    def random_link(self, state):
        """Get a random link.

        Parameters
        ----------
        state : `deque` of `str`
            Link source.

        Raises
        ------
        ValueError
            If link count is invalid.

        Returns
        -------
        (`str`, `deque` of `str`)
            Link value and updated state.
        """
        try:
            node = self.nodes[self.separator.join(state)]
            if not isinstance(node[0], list):
                state.append(node[0])
                return node[0], state
        except KeyError:
            return None, None

        nodes, values = node
        link_sum = sum(values)

        x = randint(0, link_sum - 1)
        for value, count in zip(nodes, values):
            if x < count:
                state.append(value)
                return value, state
            x -= count

        raise ValueError('invalid link sum: ' + str(self))

    @classmethod
    def load(cls, fp, override=None):
        """Load a generator.

        Parameters
        ----------
        fp : `file` or `str`
            Input file or file path.
        override : `dict` or `None`, optional
            Changes to loaded data (default: `None`).

        Returns
        -------
        `markovchain.base.MarkovBase`
            Loaded generator.
        """
        if isinstance(fp, str):
            with open(fp, 'r') as fp2:
                return cls.load(fp2, override)

        x = fp.read(1)
        fp.seek(0)

        if isinstance(x, str):
            data = json.load(fp)
        elif ijson is not None:
            try:
                data = next(ijson.items(fp, ''))
            except StopIteration:
                data = {}
        else:
            data = json.loads(fp.read().decode('utf-8'))

        if override is not None:
            extend(data, override)

        return cls(**data)

    def save(self, fp):
        """Save the generator.

        Parameters
        ----------
        fp : `file`
            Output file.
        """
        json.dump(self.get_save_data(), fp, ensure_ascii=False)

    def get_save_data(self):
        """Convert the generator to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().get_save_data()
        data['nodes'] = self.nodes
        return data



from collections import deque
from itertools import chain, repeat

from .scanner import Scanner, RegExpScanner
from .parser import ParserBase, Parser
from .util import load


class MarkovBase:
    """Base markov chain generator class.

    Attributes
    ----------
    DEFAULT_SCANNER : `type`
        Default scanner class.
    DEFAULT_PARSER : `type`
        Default parser class.
    scanner : `markovchain.scanner.Scanner`
    parser : `markovchain.parser.ParserBase`
    """
    DEFAULT_SCANNER = RegExpScanner
    DEFAULT_PARSER = Parser

    def __init__(self, separator=' ', scanner=None, parser=None):
        """Base markov chain generator constructor.

        Attributes
        ----------
        separator : `str`, optional
            State separator (default: ' ').
        scanner : `dict` or `markovchain.scanner.Scanner`, optional
            Scanner (default: `markovchain.base.MarkovBase.DEFAULT_SCANNER()`).
        parser : `dict` or `markovchain.parser.ParserBase`, optional
            Parser (default: `markovchain.base.MarkovBase.DEFAULT_PARSER()`).
        """
        self._separator = None
        self.separator = separator
        self.scanner = load(scanner, Scanner, self.DEFAULT_SCANNER)
        self.parser = load(parser, ParserBase, self.DEFAULT_PARSER)

    @property
    def separator(self):
        """`str` : State separator.
        """
        return self._separator

    @separator.setter
    def separator(self, separator):
        old_separator = self._separator
        if old_separator is not None:
            self.replace_state_separator(old_separator, separator)
        self._separator = separator

    def __eq__(self, markov):
        return (self.separator == markov.separator
                and self.scanner == markov.scanner
                and self.parser == markov.parser)

    def data(self, data, part=False):
        """Parse data and update links.

        Parameters
        ----------
        data
            Data to parse.
        part : `bool`, optional
            True if data is partial (default: `False`).
        """
        #if self.parser is None:
        #    raise ValueError('no parser')

        self.links(self.parser(self.scanner(data, part), part))

    def generate(self, maxlength, state_size=None, start=None):
        """Generate a sentence.

        Parameters
        ----------
        maxlength : `int`
            Maximum sentence length.
        state_size : `int`, optional
            State size (default: parser.state_sizes[0]).
        start : `list` of `str`, optional
            Starting state (default: []).

        Returns
        -------
        `generator` of `str`
            Word generator.
        """
        if maxlength <= 0:
            return

        if state_size is None:
            if self.parser is None:
                raise ValueError('parser is None and state_size is None')
            try:
                state_size = next(iter(self.parser.state_sizes))
            except StopIteration:
                return
        #elif (self.parser is not None
        #      and state_size not in self.parser.state_sizes):
        #    raise ValueError('invalid state size: {0}: not in {1}'
        #                     .format(state_size, self.parser.state_sizes))

        if start is None:
            start = self.separator.join(repeat('', state_size))
            state = repeat('', state_size)
        else:
            if isinstance(start, str):
                start = start.split(self.separator)
            state = chain(repeat('', state_size), start)

        state = deque(state, maxlen=state_size)

        for _ in range(maxlength):
            link, state = self.random_link(state)
            if link is None:
                return
            yield link

    def get_save_data(self):
        """Convert generator settings to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        return {
            'separator': self.separator,
            'scanner': None if self.scanner is None else self.scanner.save(),
            'parser': None if self.parser is None else self.parser.save()
        }

    def replace_state_separator(self, old_separator, new_separator):
        """Replace state separator.

        Parameters
        ----------
        old_separator : `str`
            Old state separator.
        new_separator : `str`
            New state separator.
        """
        pass

    def links(self, links):
        """Update links.

        Parameters
        ----------
        links : `generator` of (`islice` of `str`, `str`)
            Links to add.

        Raises
        -------
        NotImplementedError
            If data mixin is missing.
        """
        raise NotImplementedError('missing data mixin')

    def random_link(self, state):
        """Get a random link.

        Parameters
        ----------
        state
            Link source.

        Raises
        -------
        NotImplementedError
            If data mixin is missing.
        """
        raise NotImplementedError('missing data mixin')

    @classmethod
    def load(cls, *args, **kwargs):
        """Load a generator.

        Raises
        -------
        NotImplementedError
            If data mixin is missing.
        """
        raise NotImplementedError('missing data mixin')

    def save(self, *args, **kwargs):
        """Save the generator.

        Raises
        -------
        NotImplementedError
            If data mixin is missing.
        """
        raise NotImplementedError('missing data mixin')



from unittest import TestCase
from unittest.mock import patch

from markovchain.text.util import (
    ispunct, lstrip_ws_and_chars, capitalize,
    format_sentence_string, format_sentence
)


class TestTextUtils(TestCase):
    def test_ispunct(self):
        self.assertTrue(ispunct('\'"?,+-.[]{}()<>'))
        self.assertFalse(ispunct('\'"?,+-x.[]{}()<>'))
        self.assertFalse(ispunct(''))

    def test_capitalize(self):
        self.assertEqual(capitalize('worD WORD WoRd'), 'Word word word')
        self.assertEqual(capitalize('x'), 'X')
        self.assertEqual(capitalize(''), '')

    def test_lstrip_ws_and_chars(self):
        self.assertEqual(lstrip_ws_and_chars('', ''), '')
        self.assertEqual(lstrip_ws_and_chars('     ', ''), '')
        self.assertEqual(lstrip_ws_and_chars('  x  ', 'xy'), '')
        self.assertEqual(lstrip_ws_and_chars(' \t.\n , .x. ', '.,?!'), 'x. ')

    def test_format_sentence_string(self):
        fmt = format_sentence_string
        self.assertEqual(fmt(''), '')
        self.assertEqual(fmt('  '), '')
        self.assertEqual(fmt('  ...'), '')
        self.assertEqual(fmt('.?!word'), 'Word.')
        self.assertEqual(fmt('word', default_end='/'), 'Word/')
        self.assertEqual(fmt('word', end_chars='d'), 'Word')
        self.assertEqual(fmt('word  ,  (word)..  word'), 'Word, (word).. word.')
        self.assertEqual(fmt('word,wo[rd..wo]rd'), 'Word, wo [rd.. wo] rd.')
        self.assertEqual(fmt('wo--*--rd'), 'Wo --*-- rd.')

    @patch('markovchain.text.util.format_sentence_string', return_value=1)
    def test_format_sentence(self, fmt):
        self.assertEqual(format_sentence('word'), 1)
        fmt.assert_called_with('word', '.?!', '.')
        fmt.reset_mock()

        self.assertEqual(
            format_sentence((str(x) for x in range(3)),
                            end_chars='/[', default_end='/'),
            1
        )
        fmt.assert_called_with('0 1 2', '/[', '/')
        fmt.reset_mock()

        self.assertEqual(
            format_sentence(['a', 'b', 'c'], join_with='.'),
            1
        )
        fmt.assert_called_with('a.b.c', '.?!', '.')
        fmt.reset_mock()



import re

FORMAT_REPLACE = [(re.compile(expr), repl) for expr, repl in [
    (r'\s+'               , r' '       ), # pylint:disable=bad-whitespace
    (r'\s*([^\w\s]+)\s*'  , r'\1'      ), # pylint:disable=bad-whitespace
    (r'([,.?!])(\w)'      , r'\1 \2'   ), # pylint:disable=bad-whitespace
    (r'([\w,.?!])([[({<])', r'\1 \2'   ), # pylint:disable=bad-whitespace
    (r'([])}>])(\w)'      , r'\1 \2'   ), # pylint:disable=bad-whitespace
    (r'(\w)([-+*]+)(\w)'  , r'\1 \2 \3'), # pylint:disable=bad-whitespace
]]

RE_PUNCT = re.compile(r'^[^\w\s]+$')

def ispunct(string):
    """Return `True` if all characters in a string are punctuation
       and it is not empty.

    Parameters
    ----------
    string : `str`

    Returns
    -------
    `bool`
    """
    return RE_PUNCT.match(string) is not None

def lstrip_ws_and_chars(string, chars):
    """Remove leading whitespace and characters from a string.

    Parameters
    ----------
    string : `str`
        String to strip.
    chars : `str`
        Characters to remove.

    Returns
    -------
    `str`
        Stripped string.

    Examples
    --------
    >>> lstrip_ws_and_chars(' \\t.\\n , .x. ', '.,?!')
    'x. '
    """
    res = string.lstrip().lstrip(chars)
    while len(res) != len(string):
        string = res
        res = string.lstrip().lstrip(chars)
    return res

def capitalize(string):
    """Capitalize a sentence.

    Parameters
    ----------
    string : `str`
        String to capitalize.

    Returns
    -------
    `str`
        Capitalized string.

    Examples
    --------
    >>> capitalize('worD WORD WoRd')
    'Word word word'
    """
    if not string:
        return string
    if len(string) == 1:
        return string.upper()
    return string[0].upper() + string[1:].lower()

def format_sentence_string(string, end_chars='.?!', default_end='.'):
    """Format a sentence.

    Parameters
    ----------
    string : `str`
        Sentence.
    end_chars : `str`, optional
        Sentence ending characters (default: '.?!').
    default_end : `str`, optional
        Default sentence ending character (default: '.').

    Returns
    -------
    str
        Formatted sentence.

    Examples
    --------
    >>> format_sentence_string('  ..?!word  ,  (word)..  word')
    'Word, (word).. word.'
    >>> format_sentence_string('word WORD', default_end='?')
    'Word word?'
    >>> format_sentence_string('word,', end_chars=',')
    'Word,'
    """
    string = lstrip_ws_and_chars(string.rstrip(), end_chars)

    if not string:
        return string

    if string[-1] not in end_chars:
        string += default_end

    string = capitalize(string)

    for expr, repl in FORMAT_REPLACE:
        string = re.sub(expr, repl, string)

    return string

def format_sentence(parts, join_with=' ', end_chars='.?!', default_end='.'):
    """Format a sentence.

    Parameters
    ----------
    parts : `str` or `iterable` of `str`
        Sentence parts.
    join_with : `str`, optional
        Part separator (default: ' ').
    end_chars : `str`, optional
        Sentence ending characters (default: '.?!').
    default_end : `str`, optional
        Default sentence ending character (default: '.').

    Returns
    -------
    str
        Formatted sentence.

    Examples
    --------
    >>> format_sentence('word WORD', default_end='?')
    'Word word?'
    >>> format_sentence('word,', end_chars=',')
    'Word,'
    >>> format_sentence('word' for _ in range(3))
    'Word word word.'
    >>> format_sentence(('word' for _ in range(3)), join_with=',')
    'Word, word, word.'
    """
    if isinstance(parts, str):
        sentence = parts
    else:
        sentence = join_with.join(parts)

    return format_sentence_string(sentence, end_chars, default_end)



from unittest import TestCase
from collections import Counter
from itertools import islice

from markovchain.image.util import palette


class TestPalette(TestCase):
    def test_errors(self):
        tests = [
            (8, 8, 5),
            (0, 4, 4),
            (4, 0, 4),
            (4, 4, 0),
            (8, -4, 8),
            (-8, 4, -8)
        ]
        for test in tests:
            with self.assertRaises(ValueError):
                palette(*test)

    def test_generate(self):
        tests = [
            (8, 4, 8),
            (2, 2, 2),
            (16, 1, 1)
        ]
        for test in tests:
            res = palette(*test)
            self.assertEqual(len(res), 768)
            res = list(zip(islice(res, 0, None, 3),
                           islice(res, 1, None, 3),
                           islice(res, 2, None, 3)))
            counter = Counter(res)
            size = test[0] * test[1] * test[2]
            self.assertEqual(len(counter.items()), min(256, size + 1))
            self.assertEqual(counter[(0, 0, 0)], 256 - size)

    def test_grayscale(self):
        tests = [
            (1, 1, 1),
            (1, 1, 2),
            (1, 1, 256)
        ]
        for test in tests:
            res = palette(*test)
            self.assertEqual(len(res), 768)
            res = list(zip(islice(res, 0, None, 3),
                           islice(res, 1, None, 3),
                           islice(res, 2, None, 3)))
            self.assertTrue(all(r == g and g == b for r, g, b in res))
            counter = Counter(res)
            size = test[0] * test[1] * test[2]
            self.assertEqual(len(counter.items()), size)
            self.assertEqual(counter[(0, 0, 0)], 256 - size + 1)
            if size > 1:
                self.assertEqual(counter[(255, 255, 255)], 1)



from itertools import islice, repeat
from copy import deepcopy


class SaveLoad:
    """Base class for converting to/from JSON.

    Attributes
    ----------
    classes : `dict`
        Class group.

    Examples
    --------
    >>> class SaveLoadGroup(SaveLoad):
    ...     classes = {}
    ...
    >>> class SaveLoadObject(SaveLoadGroup):
    ...     def __init__(self, attr=None):
    ...         self.attr = attr
    ...     def save(self):
    ...         data = super().save()
    ...         data['attr'] = self.attr
    ...         return data
    ...
    >>> SaveLoadGroup.add_class(SaveLoadObject)
    >>> SaveLoadGroup.classes
    {'SaveLoadObject': <class '__main__.SaveLoadObject'>}
    >>> obj = SaveLoadObject(0)
    >>> data = obj.save()
    >>> data
    {'attr': 0, '__class__': 'SaveLoadObject'}
    >>> obj2 = SaveLoadGroup.load(data)
    >>> type(obj2)
    <class '__main__.SaveLoadObject'>
    >>> obj2.attr
    0
    """

    classes = {}

    @classmethod
    def add_class(cls, *args):
        """Add classes to the group.

        Parameters
        ----------
        *args : `type`
            Classes to add.
        """
        for cls2 in args:
            cls.classes[cls2.__name__] = cls2

    @classmethod
    def remove_class(cls, *args):
        """Remove classes from the group.

        Parameters
        ----------
        *args : `type`
            Classes to remove.
        """
        for cls2 in args:
            try:
                del cls.classes[cls2.__name__]
            except KeyError:
                pass

    @classmethod
    def load(cls, data):
        """Create an object from JSON data.

        Parameters
        ----------
        data : `dict`
            JSON data.

        Returns
        ----------
        `object`
            Created object.

        Raises
        ------
        KeyError
            If `data` does not have the '__class__' key
            or the necessary class is not in the class group.
        """
        ret = cls.classes[data['__class__']]
        data_cls = data['__class__']
        del data['__class__']
        try:
            ret = ret(**data)
        finally:
            data['__class__'] = data_cls
        return ret

    def save(self):
        """Convert an object to JSON.

        Returns
        ----------
        `dict`
            JSON data.
        """
        return {
            '__class__': self.__class__.__name__
        }


class ObjectWrapper: # pylint:disable=too-few-public-methods
    """Base class for wrapping objects.

    Example
    -------
    >>> class Object:
    ...     def method(self):
    ...         return 2
    ...
    >>> class Wrapper(ObjectWrapper):
    ...     def method(self):
    ...         return super().method() * 2
    ...
    >>> obj = Object()
    >>> wrapped = Wrapper(obj)
    >>> wrapped.method()
    4
    """
    def __init__(self, obj):
        self.__class__ = type(
            obj.__class__.__name__,
            (self.__class__, obj.__class__),
            {}
        )
        self.__dict__ = obj.__dict__


def const(x):
    """Return a function that takes any arguments and returns the specified value.

    Parameters
    ----------
    x
        Value to return.

    Returns
    -------
    `function`
    """
    return lambda *args: x

def to_list(x):
    """Convert a value to a list.

    Parameters
    ----------
    x
        Value.

    Returns
    -------
    `list`

    Examples
    --------
    >>> to_list(0)
    [0]
    >>> to_list({'x': 0})
    [{'x': 0}]
    >>> to_list(x ** 2 for x in range(3))
    [0, 1, 4]
    >>> x = [1, 2, 3]
    >>> to_list(x)
    [1, 2, 3]
    >>> _ is x
    True
    """
    if isinstance(x, list):
        return x
    if not isinstance(x, dict):
        try:
            return list(x)
        except TypeError:
            pass
    return [x]

def fill(xs, length, copy=False):
    """Convert a value to a list of specified length.

    If the input is too short, fill it with its last element.

    Parameters
    ----------
    xs
        Input list or value.
    length : `int`
        Output list length.
    copy : `bool`, optional
        Deep copy the last element to fill the list (default: False).

    Returns
    -------
    `list`

    Raises
    ------
    ValueError
        If `xs` is empty and `length` > 0

    Examples
    --------
    >>> fill(0, 3)
    [0, 0, 0]
    >>> fill((x ** 2 for x in range(3)), 1)
    [0]
    >>> x = [{'x': 0}, {'x': 1}]
    >>> y = fill(x, 4)
    >>> y
    [{'x': 0}, {'x': 1}, {'x': 1}, {'x': 1}]
    >>> y[2] is y[1]
    True
    >>> y[3] is y[2]
    True
    >>> y = fill(x, 4, True)
    >>> y
    [{'x': 0}, {'x': 1}, {'x': 1}, {'x': 1}]
    >>> y[2] is y[1]
    False
    >>> y[3] is y[2]
    False
    """
    if isinstance(xs, list) and len(xs) == length:
        return xs

    if length <= 0:
        return []

    try:
        xs = list(islice(xs, 0, length))
        if not xs:
            raise ValueError('empty input')
    except TypeError:
        xs = [xs]

    if len(xs) < length:
        if copy:
            last = xs[-1]
            xs.extend(deepcopy(last) for _ in range(length - len(xs)))
        else:
            xs.extend(islice(repeat(xs[-1]), 0, length - len(xs)))

    return xs

def load(obj, cls, default_factory):
    """Create or load an object if necessary.

    Parameters
    ----------
    obj : `object` or `dict` or `None`
    cls : `type`
    default_factory : `function`

    Returns
    -------
    `object`
    """
    if obj is None:
        return default_factory()
    if isinstance(obj, dict):
        return cls.load(obj)
    return obj

def _extend(dst, src):
    for key, val in src.items():
        if isinstance(val, dict):
            try:
                old = dst[key]
                if isinstance(old, dict):
                    _extend(old, val)
                else:
                    dst[key] = val
            except KeyError:
                dst[key] = val
        else:
            dst[key] = val

def extend(dst, *args):
    """Recursively update a dictionary.

    Parameters
    ----------
    dst : `dict`
        Dictionary to update.
    *args : `dict`

    Returns
    -------
    `dict`
        Updated dictionary.

    Examples
    --------
    >>> extend({'x': {'y': 0}}, {'x': {'z': 1}})
    {'x': {'y': 0, 'z': 1}}
    """
    for src in args:
        _extend(dst, src)
    return dst

def truncate(string, maxlen, end=True):
    """Truncate a string.

    Parameters
    ----------
    string : `str`
        String to truncate.
    maxlen : `int`
        Maximum string length.
    end : `boolean`, optional
        Remove characters from the end (default: `True`).

    Raises
    ------
    ValueError
        If `maxlen` <= 3.

    Returns
    -------
    `str`
        Truncated string.

    Examples
    --------
    >>> truncate('str', 6)
    'str'
    >>> truncate('long string', 8)
    'long ...'
    >>> truncate('long string', 8, False)
    '...tring'
    """
    if maxlen <= 3:
        raise ValueError('maxlen <= 3')

    if len(string) <= maxlen:
        return string

    if end:
        return string[:maxlen - 3] + '...'

    return '...' + string[3 - maxlen:]



from argparse import FileType
from sys import stdout #, stderr
from os import replace, remove, path, SEEK_SET, SEEK_END
from itertools import chain

from .util import format_sentence
from .. import CharScanner
from ..util import truncate
from ..cli.util import (
    load, save, infiles, JSON, SQLITE,
    tqdm, BAR_FORMAT, BAR_DESC_SIZE
)
from ..cli.util import cmd_settings # pylint:disable=unused-import

BASE = ()

def create_arg_parser(parent):
    """Create command subparsers.

    Parameters
    ----------
    parent : `argparse.ArgumentParser`
        Command parser.
    """

    arg1 = parent.add_subparsers(dest='command')

    arg2 = arg1.add_parser('create')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-o', '--output',
                      default=None,
                      help='output file (default: stdout)')
    arg2.add_argument('input', nargs='*',
                      help='input file (default: stdin)')

    arg2 = arg1.add_parser('update')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-o', '--output',
                      default=None,
                      help='output file (default: rewrite state file)')
    arg2.add_argument('state',
                      help='state file')
    arg2.add_argument('input', nargs='*',
                      help='input file (default: stdin)')

    arg2 = arg1.add_parser('settings')
    arg2.add_argument('state',
                      help='state file')

    arg2 = arg1.add_parser('generate')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-ss', '--state-size',
                      type=int, default=None,
                      help='generator state size')
    arg2.add_argument('-st', '--start',
                      default=None,
                      help='sentence start')
    arg2.add_argument('-w', '--words',
                      type=int, default=256,
                      help='max sentence size (default: %(default)s)')
    arg2.add_argument('-ws', '--word-separator',
                      default=' ',
                      help='output word separator (default: \' \')')
    arg2.add_argument('-S', '--sentences',
                      type=int, default=1,
                      help='number of generated sentences (default: %(default)s)')
    arg2.add_argument('-o', '--output',
                      type=FileType('w'), default=stdout,
                      help='output file (default: stdout)')
    arg2.add_argument('state',
                      help='state file')

def read(fnames, markov, progress):
    """Read data files and update a generator.

    Parameters
    ----------
    fnames : `list` of `str`
        File paths.
    markov : `markovchain.base.MarkovBase`
        Generator to update.
    progress : `bool`
        Show progress bar.
    """
    with infiles(fnames, progress) as fnames:
        for fname in fnames:
            with open(fname, 'r') as fp:
                if progress:
                    fp.seek(0, SEEK_END)
                    total = fp.tell()
                    title = truncate(fname, BAR_DESC_SIZE - 1, False)
                    pbar = tqdm(total=total, desc=title,
                                leave=False, unit='byte',
                                bar_format=BAR_FORMAT, dynamic_ncols=True)
                    fp.seek(0, SEEK_SET)
                    prev = 0
                else:
                    pbar = None

                try:
                    line = fp.readline()
                    while line:
                        markov.data(line.lower(), True)
                        if pbar is not None:
                            pos = fp.tell()
                            if pos <= total:
                                pbar.update(pos - prev)
                                prev = pos
                        line = fp.readline()
                finally:
                    if pbar is not None:
                        pbar.close()

                markov.data('', False)

def cmd_create(args):
    """Create a generator.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    if args.type == SQLITE:
        if args.output is not None and path.exists(args.output):
            remove(args.output)
        args.settings['db'] = args.output

    markov = args.markov(**args.settings)
    read(args.input, markov, args.progress)
    save(markov, args.output, args)

def cmd_update(args):
    """Update a generator.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    args.output = None

    markov = load(args.markov, args.state, args)
    read(args.input, markov, args.progress)
    if args.output is None:
        if args.type == SQLITE:
            save(markov, None, args)
        elif args.type == JSON:
            name, ext = path.splitext(args.state)
            tmp = name + '.tmp' + ext
            save(markov, tmp, args)
            replace(tmp, args.state)
    else:
        save(markov, args.output, args)

def cmd_generate(args):
    """Generate text.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """

    markov = load(args.markov, args.state, args)
    ss = range(args.sentences)

    if args.progress:
        title = truncate(args.output.name, BAR_DESC_SIZE - 1, False)
        ss = tqdm(ss, desc=title,
                  bar_format=BAR_FORMAT, dynamic_ncols=True)

    if args.start is not None:
        parse = markov.parser(markov.scanner(args.start.lower(), True), True)
        for _ in parse:
            pass
        state = list(markov.parser.state)
        markov.scanner.reset()
        markov.parser.reset()
    else:
        state = None

    if isinstance(markov.scanner, CharScanner):
        separator = ''
    else:
        separator = ' '

    for _ in ss:
        data = markov.generate(args.words,
                               state_size=args.state_size,
                               start=state)
        if args.start is not None:
            data = chain((args.start,), data)
        data = format_sentence(data, join_with=separator)
        if data:
            print(data)



from math import ceil, log2

from ..util import SaveLoad, load


class Traversal(SaveLoad):
    """Base image traversal class.

    Attributes
    ----------
    classes : `dict`
        Image traversal class group.
    """
    classes = {}

    def __call__(self, width, height, ends=True):
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Generate block ends (default: `True`).

        Raises
        ------
        NotImplementedError

        Returns
        -------
        `generator` of ((`int`, `int`) or `None`)
            Points and block ends.
        """
        raise NotImplementedError()


class Lines(Traversal): # pylint: disable=abstract-method
    """Base line traversal class.

    Attributes
    ----------
    reverse : `int`
        If greater than 0, reverse every nth line.
    line_sentences : `bool`
        Generate a block end after each line.
    """
    def __init__(self, reverse=0, line_sentences=False):
        """Base line traversal constructor.

        Parameters
        ----------
        reverse : `int`, optional
            If greater than 0, reverse every nth line (default: 0).
        line_sentences : `bool`, optional
            Generate a block end after each line (default: `False`).
        """
        super().__init__()
        self.reverse = reverse
        self.line_sentences = line_sentences

    def __eq__(self, t):
        return (self.reverse == t.reverse
                and self.line_sentences == t.line_sentences)

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['reverse'] = self.reverse
        data['line_sentences'] = self.line_sentences
        return data


class HLines(Lines):
    """Horizontal line traversal.

    Examples
    --------
    >>> traverse = HLines()
    >>> list(traverse(2, 2))
    [(0, 0), (1, 0), (0, 1), (1, 1)]
    >>> traverse = HLines(reverse=1)
    >>> list(traverse(2, 2))
    [(1, 0), (0, 0), (1, 1), (0, 1)]
    >>> traverse = HLines(reverse=2)
    >>> list(traverse(2, 2))
    [(1, 0), (0, 0), (0, 1), (1, 1)]
    >>> traverse = HLines(line_sentences=True)
    >>> list(traverse(2, 2))
    [(0, 0), (1, 0), None, (0, 1), (1, 1), None]
    >>> list(traverse(2, 2, ends=False))
    [(0, 0), (1, 0), (0, 1), (1, 1)]
    """
    def __call__(self, width, height, ends=True):
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Generate block ends (default: `True`).

        Returns
        -------
        `generator` of ((`int`, `int`) or `None`)
            Points and block ends.
        """
        for y in range(height):
            if self.reverse > 0 and y % self.reverse == 0:
                xs = range(width - 1, -1, -1)
            else:
                xs = range(width)

            for x in xs:
                yield x, y

            if self.line_sentences and ends:
                yield None


class VLines(Lines):
    """Vertical line traversal.

    Examples
    --------
    >>> traverse = VLines()
    >>> list(traverse(2, 2))
    [(0, 0), (0, 1), (1, 0), (1, 1)]
    >>> traverse = VLines(reverse=1)
    >>> list(traverse(2, 2))
    [(0, 1), (0, 0), (1, 1), (1, 0)]
    >>> traverse = VLines(reverse=2)
    >>> list(traverse(2, 2))
    [(1, 0), (0, 0), (1, 0), (1, 1)]
    >>> traverse = VLines(line_sentences=True)
    >>> list(traverse(2, 2))
    [(0, 0), (0, 1), None, (1, 0), (1, 1), None]
    >>> list(traverse(2, 2, ends=False))
    [(0, 0), (0, 1), (1, 0), (1, 1)]
    """
    def __call__(self, width, height, ends=True):
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Generate block ends (default: `True`).

        Returns
        -------
        generator of ((`int`, `int`) or `None`)
            Points and block ends.
        """
        for x in range(width):
            if self.reverse > 0 and x % self.reverse == 0:
                ys = range(height - 1, -1, -1)
            else:
                ys = range(height)

            for y in ys:
                yield x, y

            if self.line_sentences and ends:
                yield None


class Spiral(Traversal):
    """Spiral traversal.

    Attributes
    ----------
    reverse : `bool`
        Reverse traversal order.

    Examples
    --------
    >>> traverse = Spiral()
    >>> list(traverse(3, 3))
    [(1, 1), (0, 1), (0, 2), (1, 2), (2, 2), (2, 1), (2, 0), (1, 0), (0, 0)]
    >>> traverse = Spiral(True)
    >>> list(traverse(3, 3))
    [(0, 0), (1, 0), (2, 0), (2, 1), (2, 2), (1, 2), (0, 2), (0, 1), (1, 1)]
    """
    def __init__(self, reverse=False):
        """Spiral traversal constructor.

        Parameters
        ----------
        reverse : `bool`, optional
            Reverse traversal order (default: `False`).
        """
        super().__init__()
        self.reverse = reverse

    def __eq__(self, t):
        return self.reverse == t.reverse

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['reverse'] = self.reverse
        return data

    @staticmethod
    def _rspiral(width, height):
        """Reversed spiral generator.

        Parameters
        ----------
        width : `int`
            Spiral width.
        height : `int`
            Spiral height.

        Returns
        -------
        `generator` of (`int`, `int`)
            Points.
        """

        x0 = 0
        y0 = 0
        x1 = width - 1
        y1 = height - 1

        while x0 < x1 and y0 < y1:
            for x in range(x0, x1):
                yield x, y0
            for y in range(y0, y1):
                yield x1, y
            for x in range(x1, x0, -1):
                yield x, y1
            for y in range(y1, y0, -1):
                yield x0, y

            x0 += 1
            y0 += 1
            x1 -= 1
            y1 -= 1

        if x0 == x1:
            for y in range(y0, y1 + 1):
                yield x0, y
        elif y0 == y1:
            for x in range(x0, x1 + 1):
                yield x, y0

    @staticmethod
    def _spiral(width, height):
        """Spiral generator.

        Parameters
        ----------
        width : `int`
            Spiral width.
        height : `int`
            Spiral height.

        Returns
        -------
        `generator` of (`int`, `int`)
            Points.
        """

        if width == 1:
            for y in range(height - 1, -1, -1):
                yield 0, y
            return
        if height == 1:
            for x in range(width - 1, -1, -1):
                yield x, 0
            return

        if width <= height:
            x0 = width // 2
            if width % 2:
                for y in range(height - 1 - x0, x0 - 1, -1):
                    yield x0, y
            x0 -= 1
            y0 = x0
        else:
            y0 = height // 2
            if height % 2:
                for x in range(width - 1 - y0, y0 - 1, -1):
                    yield x, y0
            y0 -= 1
            x0 = y0

        while x0 >= 0:
            x1 = width - x0 - 1
            y1 = height - y0 - 1

            for y in range(y0 + 1, y1):
                yield x0, y
            for x in range(x0, x1):
                yield x, y1
            for y in range(y1, y0, -1):
                yield x1, y
            for x in range(x1, x0 - 1, -1):
                yield x, y0

            x0 -= 1
            y0 -= 1

    def __call__(self, width, height, ends=True): # pylint: disable=unused-argument
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Unused (default: `True`).

        Returns
        -------
        `generator` of (`int`, `int`)
            Points.
        """
        if self.reverse:
            yield from self._rspiral(width, height)
        else:
            yield from self._spiral(width, height)


class Hilbert(Traversal):
    """Hilbert curve traversal.

    Attributes
    ----------
    POSITION : `list` of (`int`, `int`)
        Block positions.

    Examples
    --------
    >>> traverse = Hilbert()
    >>> list(traverse(2, 2))
    [(0, 0), (0, 1), (1, 1), (1, 0)]
    >>> list(traverse(3, 5))
    [(0, 0), (0, 1), (1, 1), (1, 0), (2, 0), (2, 1),
     (2, 2), (2, 3), (1, 3), (1, 2), (0, 2), (0, 3),
     (0, 4), (1, 4), (2, 4)]
    """

    POSITION = [(0, 0), (0, 1), (1, 1), (1, 0)]

    @classmethod
    def get_point_in_block(cls, x, y, block_idx, block_size):
        """Get point coordinates in next block.

        Parameters
        ----------
        x : `int`
            X coordinate in current block.
        y : `int`
            Y coordinate in current block.
        block_index : `int`
            Current block index in next block.
        block_size : `int`
            Current block size.

        Raises
        ------
        IndexError
            If block index is out of range.

        Returns
        -------
        (`int`, `int`)
            Point coordinates.
        """
        if block_idx == 0:
            return y, x
        if block_idx == 1:
            return x, y + block_size
        if block_idx == 2:
            return x + block_size, y + block_size
        if block_idx == 3:
            x, y = block_size - 1 - y, block_size - 1 - x
            return x + block_size, y
        raise IndexError('block index out of range: %d' % block_idx)

    @classmethod
    def get_point(cls, idx, size):
        """Get curve point coordinates by index.

        Parameters
        ----------
        idx : `int`
            Point index.
        size : `int`
            Bounding box size.

        Returns
        -------
        (`int`, `int`)
            Point coordinates.
        """
        x, y = cls.POSITION[idx % 4]
        idx //= 4
        block_size = 2
        while block_size < size:
            block_idx = idx % 4
            x, y = cls.get_point_in_block(x, y, block_idx, block_size)
            idx //= 4
            block_size *= 2
        return x, y

    def __call__(self, width, height, ends=True): # pylint: disable=unused-argument
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Unused (default: `True`).

        Returns
        -------
        `generator` of (`int`, `int`)
            Points.
        """
        size = max(width, height)
        size = 2 ** ceil(log2(size))

        generated = 0
        points = width * height
        if generated >= points or width <= 0 or height <= 0:
            return

        for i in range(size * size):
            x, y = self.get_point(i, size)
            if x < width and y < height:
                yield x, y
                generated += 1
                if generated >= points:
                    return

    def __eq__(self, tr):
        return isinstance(tr, self.__class__)


class Blocks(Traversal):
    """Block traversal.

    Attributes
    ----------
    block_size : (`int`, `int`)
        Block size.
    block_sentences : `bool`
        Generate a block end after each block.
    traverse_image : `markovchain.image.traversal.Traversal`
        Image traversal.
    traverse_block : `markovchain.image.traversal.Traversal`
        Block traversal.

    Examples
    --------
    >>> traverse = Blocks(block_size=(2, 2),
    ...                   traverse_image=HLines(),
    ...                   traverse_block=VLines())
    >>> list(traverse(4, 4))
    [(0, 0), (0, 1), (1, 0), (1, 1),
     (2, 0), (2, 1), (3, 0), (3, 1),
     (0, 2), (0, 3), (1, 2), (1, 3),
     (2, 2), (2, 3), (3, 2), (3, 3)]
    """
    def __init__(self,
                 block_size=(16, 16),
                 block_sentences=False,
                 traverse_image=None,
                 traverse_block=None):
        """ Block traversal constructor.

        Parameters
        ----------
        block_size : (`int`, `int`), optional
            Block size (default: (16, 16)).
        block_sentences : `bool`, optional
            Generate a block end after each block (default: False).
        traverse_image : `markovchain.image.traversal.Traversal` \
                          or `dict`, optional
            Image traversal object or JSON data (default: HLines()).
        traverse_block : `markovchain.image.traversal.Traversal` \
                         or `dict`, optional
            Block traversal object or JSON data (default: HLines()).
        """
        super().__init__()
        self.block_size = tuple(block_size)
        self.block_sentences = block_sentences
        self.traverse_image = load(traverse_image, Traversal, HLines)
        self.traverse_block = load(traverse_block, Traversal, HLines)

    def __call__(self, width, height, ends=True):
        """Traverse an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        ends : `bool`, optional
            Generate block ends (default: `True`).

        Returns
        -------
        `generator` of ((`int`, `int`) or `None`)
            Points and block ends.
        """
        block_width, block_height = self.block_size
        img_width = width // block_width
        img_height = height // block_height
        for xy in self.traverse_image(img_width, img_height, ends):
            if xy is None:
                yield None
                continue
            x, y = xy
            x, y = x * block_width, y * block_height
            for dx, dy in self.traverse_block(block_width, block_height, False):
                yield x + dx, y + dy
            if self.block_sentences and ends:
                yield None

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['block_size'] = list(self.block_size)
        data['block_sentences'] = self.block_sentences
        data['traverse_image'] = self.traverse_image.save()
        data['traverse_block'] = self.traverse_block.save()
        return data

    def __eq__(self, t):
        return (self.block_size == t.block_size
                and self.block_sentences == t.block_sentences
                and self.traverse_image == t.traverse_image
                and self.traverse_block == t.traverse_block)



from unittest import TestCase
from itertools import product

from markovchain.image.traversal import (
    Traversal, Lines, HLines, VLines, Spiral, Hilbert, Blocks
)


class TestLines(TestCase):
    def setUp(self):
        Traversal.add_class(Lines)

    def tearDown(self):
        Traversal.remove_class(Lines)

    def test_save_load(self):
        tests = [(0, False), (2, True)]
        for test in tests:
            test = Lines(*test)
            saved = test.save()
            loaded = Traversal.load(saved)
            self.assertEqual(test, loaded)


class TestHLines(TestCase):
    def test_traverse(self):
        test = HLines()

        test.reverse = 0
        self.assertEqual(list(test(2, 3, False)),
                         [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)])
        self.assertEqual(list(test(2, 3, True)),
                         [(0, 0), (1, 0), (0, 1), (1, 1), (0, 2), (1, 2)])
        test.reverse = 1
        self.assertEqual(list(test(2, 3)),
                         [(1, 0), (0, 0), (1, 1), (0, 1), (1, 2), (0, 2)])
        test.reverse = 2
        self.assertEqual(list(test(2, 3)),
                         [(1, 0), (0, 0), (0, 1), (1, 1), (1, 2), (0, 2)])
        test.reverse = 0
        test.line_sentences = True
        self.assertEqual(list(test(2, 3, True)),
                         [(0, 0), (1, 0), None,
                          (0, 1), (1, 1), None,
                          (0, 2), (1, 2), None])
        self.assertEqual(list(test(2, 3, False)),
                         [(0, 0), (1, 0), (0, 1),
                          (1, 1), (0, 2), (1, 2)])

    def test_save_load(self):
        tests = [(0, False), (2, True)]
        for test in tests:
            test = HLines(*test)
            saved = test.save()
            loaded = Traversal.load(saved)
            self.assertEqual(test, loaded)


class TestVLines(TestCase):
    def test_traverse(self):
        test = VLines()

        test.reverse = 0
        self.assertEqual(list(test(2, 3)),
                         [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)])
        test.reverse = 1
        self.assertEqual(list(test(2, 3)),
                         [(0, 2), (0, 1), (0, 0), (1, 2), (1, 1), (1, 0)])
        test.reverse = 2
        self.assertEqual(list(test(2, 3)),
                         [(0, 2), (0, 1), (0, 0), (1, 0), (1, 1), (1, 2)])
        test.reverse = 0
        test.line_sentences = True
        self.assertEqual(list(test(2, 3)),
                         [(0, 0), (0, 1), (0, 2), None,
                          (1, 0), (1, 1), (1, 2), None])
        self.assertEqual(list(test(2, 3, False)),
                         [(0, 0), (0, 1), (0, 2),
                          (1, 0), (1, 1), (1, 2)])

    def test_save_load(self):
        tests = [(0, False), (2, True)]
        for test in tests:
            test = VLines(*test)
            saved = test.save()
            loaded = Traversal.load(saved)
            self.assertEqual(test, loaded)


class TestSpiral(TestCase):
    def test_traverse(self):
        test = Spiral()
        tests = product(range(1, 7), range(1, 7))
        for width, height in tests:
            test.reverse = False
            res = list(test(width, height, False))
            self.assertCountEqual(res, product(range(width), range(height)))
            test.reverse = True
            res2 = list(test(width, height, False))
            res.reverse()
            self.assertEqual(res2, res)

    def test_save_load(self):
        tests = [(False,), (True,)]
        for test in tests:
            test = Spiral(*test)
            saved = test.save()
            loaded = Traversal.load(saved)
            self.assertEqual(test, loaded)


class TestHilbert(TestCase):
    def test_traverse(self):
        test = Hilbert()
        tests = product(range(1, 7), range(1, 7))
        for width, height in tests:
            res = list(test(width, height))
            self.assertCountEqual(res, product(range(width), range(height)))

    def test_save_load(self):
        test = Hilbert()
        saved = test.save()
        loaded = Traversal.load(saved)
        self.assertEqual(test, loaded)


class TestBlocks(TestCase):
    def test_traverse(self):
        def hline(width, height, ends): # pylint: disable=unused-argument
            for x in range(width):
                yield (x, 0)
                if ends and x == 0:
                    yield None

        def vline(width, height, ends): # pylint: disable=unused-argument
            for y in range(height):
                yield (0, y)
                if ends:
                    yield None

        traverse = Blocks(block_size=(2, 2),
                          block_sentences=False,
                          traverse_image=hline,
                          traverse_block=vline)

        tests = [
            ((4, 4, False), [(0, 0), (0, 1), (2, 0), (2, 1)]),
            ((3, 3, False), [(0, 0), (0, 1)]),
            ((4, 4, True), [(0, 0), (0, 1), None, (2, 0), (2, 1)])
        ]

        for test, res in tests:
            self.assertEqual(list(traverse(*test)), res)

        traverse.block_sentences = True

        tests = [
            ((4, 4, False), [(0, 0), (0, 1), (2, 0), (2, 1)]),
            ((3, 3, True), [(0, 0), (0, 1), None, None]),
            ((4, 4, True), [(0, 0), (0, 1), None, None, (2, 0), (2, 1), None])
        ]

        for test, res in tests:
            self.assertEqual(list(traverse(*test)), res)

    def test_save_load(self):
        tests = [
            ((8, 8), True,
             HLines(reverse=1, line_sentences=True),
             VLines(reverse=2, line_sentences=False))
        ]
        for test in tests:
            test = Blocks(*test)
            saved = test.save()
            loaded = Traversal.load(saved)
            self.assertEqual(test, loaded)



from .scanner import Scanner, CharScanner, RegExpScanner
from .parser import Parser, LevelParser
from .base import MarkovBase
from .json import MarkovJsonMixin
from .sqlite import MarkovSqliteMixin

Parser.add_class(Parser, LevelParser)
Scanner.add_class(CharScanner, RegExpScanner)



from colorsys import hsv_to_rgb
from math import floor


def palette(hues, saturations, values):
    """Generate a palette.

    Parameters
    ----------
    hues : `int`
        Number of hues.
    saturations : `int`
        Number of saturations.
    values : `int`
        Number of values.

    Raises
    ------
    ValueError
        If `hues` * `saturations` * `values` > 256
        or min(`hues`, `saturations`, `values`) < 1.

    Returns
    -------
    `list` of `int`
        Palette for `PIL.Image.putpalette`.
    """
    size = hues * saturations * values
    if size > 256:
        raise ValueError('palette size > 256: {0}'.format(size))
    if min(hues, saturations, values) < 1:
        raise ValueError('invalid palette size: {0} {1} {2}'
                         .format(hues, saturations, values))

    ret = []
    if hues == 1 and saturations == 1:
        if values == 1:
            size = 0
        else:
            nvalues = values - 1
            for value in range(values):
                value1 = value * 255 // nvalues
                ret.extend((value1, value1, value1))
    else:
        for saturation in range(1, saturations + 1):
            saturation1 = saturation / saturations
            for hue in range(1, hues + 1):
                hue1 = hue / hues
                for value in range(1, values + 1):
                    value1 = value / values
                    ret.extend(floor(x * 255)
                               for x in hsv_to_rgb(hue1, saturation1, value1))

    ret.extend(0 for _ in range((256 - size) * 3))
    return ret


def convert(ctype, img, palette_img, dither=False):
    """Convert an image to palette type.

    Parameters
    ----------
    ctype : `int`
        Conversion type.
    img : `PIL.Image`
        Image to convert.
    palette_img : `PIL.Image`
        Palette source image.
    dither : `bool`, optional
        Enable dithering (default: `False`).

    Raises
    ------
    ValueError
        If palette_img has no palette.

    Returns
    -------
    `PIL.Image`
        Converted image.
    """
    if ctype == 0:
        img2 = img.convert(mode='P')
        img2.putpalette(palette_img.getpalette())
        return img2

    img.load()
    palette_img.load()
    if palette_img.palette is None:
        raise ValueError('invalid palette image')
    im = img.im.convert('P', int(dither), palette_img.im)
    return img._new(im) # pylint: disable=protected-access



from argparse import FileType
from sys import stderr
from os import replace, remove, path
from shutil import copyfile
from functools import reduce
from itertools import islice
from PIL import Image

from . import ImageScanner, MarkovImageMixin
from .util import palette
from ..util import ObjectWrapper, truncate
from ..cli.util import (
    tqdm, load, save, infiles, outfiles as _outfiles,
    check_output_format, JSON, SQLITE,
    BAR_FORMAT, BAR_DESC_SIZE
)
from ..cli.util import cmd_settings # pylint:disable=unused-import


BASE = (MarkovImageMixin,)


def create_arg_parser(parent):
    """Create command subparsers.

    Parameters
    ----------
    parent : `argparse.ArgumentParser`
        Command parser.
    """

    arg1 = parent.add_subparsers(dest='command')

    arg2 = arg1.add_parser('convert')
    arg2.add_argument('-p', '--palette',
                      nargs=3, metavar=('HUES', 'SATURATIONS', 'VALUES'),
                      type=int, default=[8, 4, 8],
                      help='palette color division (default: %(default)s)')
    arg2.add_argument('-c', '--convert-type',
                      type=int, default=1, choices=(0, 1),
                      help='conversion type (default: %(default)s)')
    arg2.add_argument('-d', '--dither',
                      action='store_true',
                      help='enable dithering')
    arg2.add_argument('-r', '--resize', metavar=('WIDTH', 'HEIGHT'),
                      type=int, nargs=2,
                      default=None,
                      help='resize images (default: %(default)s)')
    arg2.add_argument('input', nargs='+',
                      help='input file')

    arg2 = arg1.add_parser('create')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-o', '--output',
                      default=None,
                      help='output file (default: stdout)')
    arg2.add_argument('input', nargs='*',
                      help='input file')

    arg2 = arg1.add_parser('update')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-o', '--output',
                      default=None,
                      help='output file (default: rewrite state file)')
    arg2.add_argument('state',
                      help='state file')
    arg2.add_argument('input', nargs='*',
                      help='input file')

    arg2 = arg1.add_parser('settings')
    arg2.add_argument('state',
                      help='state file')

    arg2 = arg1.add_parser('generate')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-ss', '--state-size',
                      type=int, nargs='+', default=None,
                      help='generator state sizes')
    arg2.add_argument('-S', '--size', metavar=('WIDTH', 'HEIGHT'),
                      type=int, nargs=2, default=None,
                      help='image size (default: <scanner.resize>)')
    arg2.add_argument('-l', '--level',
                      type=int, default=None,
                      help='image levels (default: <scanner.levels>)')
    arg2.add_argument('-c', '--count',
                      type=int, default=1,
                      help='generated image count (default: %(default)s)')
    arg2.add_argument('state',
                      help='state file')
    arg2.add_argument('output',
                      help='output file name format string')

    arg2 = arg1.add_parser('filter')
    arg2.add_argument('-P', '--progress',
                      action='store_true',
                      help='show progress bar')
    arg2.add_argument('-t', '--type',
                      choices=('json', 'sqlite'), default='json',
                      help='generator type (default: %(default)s)')
    arg2.add_argument('-s', '--settings',
                      type=FileType('r'), default=None,
                      help='settings json file')
    arg2.add_argument('-S', '--state',
                      default=None,
                      help='state file')
    arg2.add_argument('-ss', '--state-size',
                      type=int, nargs='+', default=None,
                      help='generator state sizes')
    arg2.add_argument('-l', '--level',
                      type=int, default=1,
                      help='filter start level (default: %(default)s)')
    arg2.add_argument('-c', '--count',
                      type=int, default=1,
                      help='generated image count (default: %(default)s)')
    arg2.add_argument('input',
                      help='input image')
    arg2.add_argument('output',
                      help='output file name format string')


class TraversalProgressWrapper(ObjectWrapper): # pylint: disable=too-few-public-methods
    """Traversal object wrapper.

    Shows image traversal progress.

    Attributes
    ----------
    pbar_parent : `tqdm.tqdm`
        Parent progress bar.
    """
    def __init__(self, obj, parent=None):
        super().__init__(obj)
        self.pbar_parent = parent

    def __call__(self, width, height, ends=True):
        size = width * height
        if self.pbar_parent is not None:
            title = 'Level ' + str(self.pbar_parent.n + 1)
        else:
            title = None
        pbar = tqdm(total=size, desc=title,
                    leave=False, unit='px',
                    bar_format=BAR_FORMAT, dynamic_ncols=True)
        try:
            for xy in super().__call__(width, height, ends):
                if xy is not None:
                    pbar.update(1)
                yield xy
        finally:
            pbar.close()

        if self.pbar_parent is not None:
            self.pbar_parent.update(1)


def read(fnames, markov, progress, leave=True):
    """Read data files and update a generator.

    Parameters
    ----------
    fnames : `list` of `str`
        File paths.
    markov : `markovchain.base.MarkovBase`
        Generator to update.
    progress : `bool`
        Show progress bar.
    leave : `bool`, optional
        Leave progress bars (default: `True`).
    """
    tr = markov.scanner.traversal
    if progress and not isinstance(tr[0], TraversalProgressWrapper):
        tr[0] = TraversalProgressWrapper(tr[0])
    tr = tr[0]

    with infiles(fnames, progress, leave) as fnames:
        for fname in fnames:
            if progress:
                title = truncate(fname, BAR_DESC_SIZE - 1, False)
                pbar = tqdm(total=markov.levels, desc=title,
                            leave=False, unit='lvl',
                            bar_format=BAR_FORMAT, dynamic_ncols=True)
                tr.pbar_parent = pbar
            markov.data(Image.open(fname), False)
            if progress:
                pbar.close()

def outfiles(markov, fmt, nfiles, progress, start=0):
    """Get output file paths.

    Parameters
    ----------
    markov : `markovchain.base.MarkovBase`
        Markov chain generator.
    fmt : `str`
        File path format string.
    nfiles : `int`
        Number of files.
    progress : `bool`
        Show progress bars.
    start : `int`, optional
        Initial image level (default: 0).

    Returns
    -------
    `generator` of `str`
        Output file paths.
    """
    tr = markov.scanner.traversal
    if progress and not isinstance(tr[0], TraversalProgressWrapper):
        tr[0] = TraversalProgressWrapper(tr[0])
    tr = tr[0]

    with _outfiles(fmt, nfiles, progress) as fnames:
        for fname in fnames:
            if progress:
                title = truncate(fname, BAR_DESC_SIZE - 1, False)
                pbar = tqdm(initial=start, total=markov.levels,
                            desc=title, leave=False, unit='lvl',
                            bar_format=BAR_FORMAT, dynamic_ncols=True)
                tr.pbar_parent = pbar
            yield fname
            if progress:
                pbar.close()

def cmd_convert(args):
    """Convert an image.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    scanner = ImageScanner(
        levels=1,
        resize=tuple(args.resize) if args.resize else None,
        dither=args.dither,
        convert_type=args.convert_type,
        palette=palette(*args.palette)
    )

    for fname in args.input:
        name, _ = path.splitext(fname)
        name += '_convert.png'
        img = scanner.input(Image.open(fname))
        img = scanner.level(img, 0)
        with open(name, 'wb') as fp:
            img.save(fp)

def cmd_create(args):
    """Create a generator.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    if args.type == SQLITE:
        if path.exists(args.output):
            remove(args.output)
        args.settings['db'] = args.output

    markov = args.markov(**args.settings)
    read(args.input, markov, args.progress)
    save(markov, args.output, args)

def cmd_update(args):
    """Update a generator.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    if args.type == SQLITE and args.output is not None:
        copyfile(args.state, args.output)
        args.state = args.output

    markov = load(args.markov, args.state, args)

    read(args.input, markov, args.progress)

    if args.output is None:
        if args.type == SQLITE:
            save(markov, None, args)
        elif args.type == JSON:
            name, ext = path.splitext(args.state)
            tmp = name + '.tmp' + ext
            save(markov, tmp, args)
            replace(tmp, args.state)
    else:
        save(markov, args.output, args)

def cmd_generate(args):
    """Generate images.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    check_output_format(args.output, args.count)

    markov = load(args.markov, args.state, args)

    if args.size is None:
        if markov.scanner.resize is None:
            print('Unknown output image size', file=stderr)
            exit(1)
        width, height = markov.scanner.resize
    else:
        width, height = args.size

    if args.level is None:
        scale = markov.scanner.min_size
    else:
        scale = reduce(
            lambda x, y: x * y,
            islice(markov.scanner.level_scale, 0, args.level - 1),
            1
        )

    width, height = width // scale, height // scale

    markov.scanner.traversal[0].show_progress = args.progress

    for fname in outfiles(markov, args.output, args.count, args.progress):
        img = markov.image(width, height,
                           state_size=args.state_size,
                           levels=args.level)
        with open(fname, 'wb') as fp:
            img.save(fp)

def cmd_filter(args):
    """Filter an image.

    Parameters
    ----------
    args : `argparse.Namespace`
        Command arguments.
    """
    check_output_format(args.output, args.count)

    img = Image.open(args.input)
    width, height = img.size

    if args.state is not None:
        markov = load(args.markov, args.state, args)
    else:
        markov = args.markov(**args.settings)

    if args.state is None:
        read([args.input], markov, args.progress, False)

    args.level = min(args.level, markov.levels - 1) - 1

    if args.level < 0:
        args.level = -1
        scale = markov.scanner.min_size
        width, height = width // scale, height // scale
        start = None
    else:
        scale = reduce(
            lambda x, y: x * y,
            islice(markov.scanner.level_scale, args.level, markov.levels),
            1
        )
        width, height = width // scale, height // scale
        start = img.resize((width, height), markov.scanner.scale)

    for fname in outfiles(markov, args.output,
                          args.count, args.progress, args.level + 1):
        img = markov.image(width, height,
                           state_size=args.state_size,
                           start_level=args.level,
                           start_image=start)
        with open(fname, 'wb') as fp:
            img.save(fp)



from .markov import MarkovImageMixin
from .scanner import ImageScanner
from .traversal import Traversal, HLines, VLines, Spiral, Blocks, Hilbert

from ..scanner import Scanner
from ..parser import Parser, LevelParser

Scanner.add_class(ImageScanner)
Traversal.add_class(HLines, VLines, Spiral, Blocks, Hilbert)



from collections import deque
from itertools import repeat, islice

from .util import SaveLoad, to_list, fill, load
from .scanner import Scanner


class ParserBase(SaveLoad):
    """Base parser class.

    Attributes
    ----------
    classes : `dict`
        Parser class group.

    Examples
    --------
    >>> parse = ParserBase(lambda data: zip(data, data[1:]))
    >>> list(parse(['a', 'b', 'c']))
    [('a', 'b'), ('b', 'c')]
    """

    classes = {}

    def __init__(self, parse=None):
        """Base parser constructor.

        Parameters
        ----------
        parse : `function`, optional
        """
        if parse is not None:
            self.parse = parse

    def __call__(self, data, part=False):
        """Parse data.

        Parameters
        ----------
        data
            Data to parse.
        part : `bool`, optional
            True if data is partial (default: `False`).

        Returns
        -------
        `object`
            self.parse(data)
        """
        return self.parse(data)


class Parser(ParserBase):
    """Default parser class.

    Attributes
    ----------
    state : `deque` of `str`
        Parser state.
    state_size : `int`
        Maximum parser state size.
    reset_on_sentence_end : `bool`
        Reset parser state on `markovchain.scanner.Scanner.END` token.
    end : `bool`
        True if a sentence is not started.

    Examples
    --------
    >>> tokens = ['a', 'b', '.', Scanner.END, 'c', '.', Scanner.END]
    >>> parse = Parser()
    >>> [(list(state), next) for state, next in parse(tokens)]
    [([''], 'a'), (['a'], 'b'), (['b'], '.'),
     ([''], 'c'), (['c'], '.')]
    >>> parse.state_sizes = 2
    >>> [(list(state), next) for state, next in parse(tokens)]
    [(['', ''], 'a'), (['', 'a'], 'b'), (['a', 'b'], '.'),
     (['', ''], 'c'), (['', 'c'], '.')]
    >>> parse.state_sizes = [1, 2]
    >>> [(list(state), next) for state, next in parse(tokens)]
    [([''], 'a'), (['', ''], 'a'),
     (['a'], 'b'), (['', 'a'], 'b'),
     (['b'], '.'), (['a', 'b'], '.'),
     ([''], 'c'), (['', ''], 'c'),
     (['c'], '.'), (['', 'c'], '.')]
    """
    def __init__(self, state_sizes=None,
                 reset_on_sentence_end=True):
        """Default parser constructor.

        Parameters
        ----------
        state_sizes : `int` or `list` of `int`, optional
            Parser state size(s) (default: [1]).
        reset_on_sentence_end : `bool`, optional
            Reset parser state on `markovchain.scanner.Scanner.END` token (default: `True`).
        """
        super().__init__()
        self.state = None
        self.state_size = None
        self._state_sizes = None
        self.reset_on_sentence_end = reset_on_sentence_end
        self.end = True
        self.state_sizes = [1] if state_sizes is None else state_sizes

    @property
    def state_sizes(self):
        """`list` of `int` : Parser state sizes.
        """
        return self._state_sizes

    @state_sizes.setter
    def state_sizes(self, state_sizes):
        state_sizes = to_list(state_sizes)
        if self.state is None or self._state_sizes != state_sizes:
            if any(s <= 0 for s in state_sizes):
                raise ValueError('parser state size <= 0: {0}'
                                 .format(state_sizes))
            self.state_size = max(state_sizes)
            self._state_sizes = state_sizes
            self.reset(True)

    def reset(self, state_size_changed=False):
        """Reset parser state.

        Parameters
        ----------
        state_size_changed : `bool`, optional
            `True` if maximum state size changed (default: `False`).
        """
        if state_size_changed:
            self.state = deque(repeat('', self.state_size),
                               maxlen=self.state_size)
        else:
            self.state.extend(repeat('', self.state_size))
        self.end = True

    def __call__(self, data, part=False):
        """Parse tokens.

        Parameters
        ----------
        data : `generator` of (`str` or `markovchain.scanner.Scanner.END` or (`markovchain.scanner.Scanner.START`, `str`))
            Tokens to parse.
        part : `bool`, optional
            `True` if data is partial (default: `False`).

        Returns
        -------
        generator of (`islice` of `str`, `str`)
            Link generator.
        """
        for word in data:
            if isinstance(word, tuple):
                cmd, arg = word
                if cmd == Scanner.START:
                    self.reset()
                    self.state.append(arg)
                    self.end = False
                else:
                    raise ValueError('invalid parser input: {0}'.format(word))
            elif word == Scanner.END:
                if not self.end and self.reset_on_sentence_end:
                    self.reset()
                self.end = True
            else:
                for sz in self.state_sizes:
                    start = self.state_size - sz
                    yield islice(self.state, start, self.state_size), word
                self.state.append(word)
                self.end = False

        if not part:
            self.reset()

    def __eq__(self, parser):
        return (self.reset_on_sentence_end == parser.reset_on_sentence_end
                and self.state_size == parser.state_size)

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['state_sizes'] = self.state_sizes
        data['reset_on_sentence_end'] = self.reset_on_sentence_end
        return data


class LevelParser(ParserBase):
    """Multilevel parser class.
    """
    def __init__(self, levels=1, parsers=None):
        """Multilevel parser constructor.

        Parameters
        ----------
        levels : `int`
            Number of levels.
        parsers : `list` of `markovchain.parser.ParserBase`
            Level parsers.
        """
        super().__init__()
        self._parsers = None
        self._parsers_max = None
        self._levels = None
        self.levels = levels
        self.parsers = parsers

    @property
    def parsers(self):
        """`list` of `markovchain.parser.ParserBase` : Level parsers.
        """
        return self._parsers

    @parsers.setter
    def parsers(self, parsers):
        parsers = to_list(parsers)

        for i, parser in enumerate(parsers):
            parsers[i] = load(parser, Parser, Parser)

        if self.levels is not None:
            filled = fill(parsers, self.levels, True)
        else:
            filled = parsers

        self._parsers_max = parsers if len(parsers) > len(filled) else filled
        self._parsers = filled

    @property
    def levels(self):
        """`int` : Number of levels.
        """
        return self._levels

    @levels.setter
    def levels(self, levels):
        if levels <= 0:
            raise ValueError('level count <= 0')
        self._levels = levels
        if self._parsers_max is not None:
            self.parsers = self._parsers_max

    def reset(self):
        """Reset parser state.
        """
        for parser in self.parsers:
            parser.reset()

    def __call__(self, data, part=False):
        """Parse tokens.

        Parameters
        ----------
        data : `generator` of `generator` of (`str` or `markovchain.scanner.Scanner.END` or (`markovchain.scanner.Scanner.START`, `str`))
            Levels to parse.
        part : `bool`, optional
            `True` if data is partial (default: `False`).

        Returns
        -------
        `generator` of (`islice` of `str`, `str`)
            Link generator.
        """
        #if part:
        #    raise NotImplementedError()
        for level, parser in zip(data, self.parsers):
            yield from parser(level, part)

    def __eq__(self, parser):
        return (self.levels == parser.levels
                and self.parsers == parser.parsers)

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['levels'] = self.levels
        if self.parsers is None:
            data['parsers'] = None
        else:
            data['parsers'] = [parser.save() for parser in self.parsers]
        return data



from sys import argv, stderr
from argparse import ArgumentParser

from .util import set_args
from ..text import cli as text

try:
    from ..image import cli as image
except ImportError:
    image = None

def main():
    """CLI main function.
    """
    parser = ArgumentParser()
    parsers = parser.add_subparsers(dest='dtype')

    text.create_arg_parser(parsers.add_parser('text'))
    if image is not None:
        image.create_arg_parser(parsers.add_parser('image'))

    if len(argv) == 1:
        parser.print_help()
        exit(1)

    args = parser.parse_args()
    dtype = globals()[args.dtype]
    try:
        set_args(args, dtype.BASE)
        cmd = getattr(dtype, 'cmd_' + args.command)
        cmd(args)
    except ValueError as err:
        print(str(err), file=stderr)
        exit(1)



from unittest import TestCase
from unittest.mock import Mock, MagicMock, mock_open, patch

from io import StringIO
from argparse import Namespace
import json
import sys

from markovchain.cli.util import (
    no_tqdm, NoProgressBar,
    pprint,
    load, save, IJSON_MIN_SIZE, IJSON_MIN_COMPRESSED_SIZE,
    set_args, JSON, SQLITE,
    check_output_format,
    infiles, outfiles
)
from markovchain import MarkovBase, MarkovJsonMixin, MarkovSqliteMixin

class TestNoProgressBar(TestCase):
    @patch('sys.stderr', new_callable=StringIO)
    def testWarning(self, stderr):
        p = NoProgressBar()
        self.assertEqual(stderr.getvalue(), '')
        self.assertFalse(p.warning)
        NoProgressBar.print_warning()
        msg = stderr.getvalue()
        self.assertNotEqual(msg, '')
        self.assertTrue(p.warning)
        NoProgressBar.print_warning()
        self.assertEqual(stderr.getvalue(), msg)
        self.assertTrue(p.warning)

    @patch('markovchain.cli.util.NoProgressBar')
    def testNoTqdm(self, no_pbar):
        p = no_tqdm(total=100, leave=False, desc='')
        self.assertIsInstance(p, Mock)
        no_pbar.assert_called_with()
        no_pbar.print_warning.assert_called_with()
        no_pbar.reset_mock()
        it = range(3)
        p = no_tqdm(it)
        self.assertIs(p, it)
        no_pbar.assert_not_called()
        no_pbar.print_warning.assert_called_with()

class TestPPrint(TestCase):
    @patch('sys.stdout', new_callable=StringIO)
    def testPPrint(self, stdout):
        tests = [
            {'x': 0, 'y': [{'z': '0'}, {'z': '1'}]},
            [{'x': 0.1, 'y': {'y': False, 'z': None, 'u': 3}}, {'z': 0}]
        ]
        for test in tests:
            stdout.seek(0)
            stdout.truncate(0)
            pprint(test)
            self.assertEqual(
                stdout.getvalue(),
                json.dumps(test, indent=4, sort_keys=True) + '\n'
            )

    @patch('sys.stdout', new_callable=StringIO)
    def testPPrintArray(self, stdout):
        tests = [
            ([], '[]\n'),
            ([0, 1, 2], '[0, 1, 2]\n'),
            ([True, None], '[true, null]\n'),
            ([0, [1, 2], 3], '[\n    0,\n    [1, 2],\n    3\n]\n')
        ]
        for test, res in tests:
            stdout.seek(0)
            stdout.truncate(0)
            pprint(test)
            self.assertEqual(stdout.getvalue(), res)

class TestLoad(TestCase):
    @patch('builtins.open', new_callable=mock_open)
    @patch('bz2.open', new_callable=mock_open)
    @patch('os.path.getsize', return_value=IJSON_MIN_SIZE - 1)
    @patch('sys.stdout', new_callable=StringIO)
    def testOpen(self, stdout, getsize, bz2op, op):
        class Test:
            pass
        Test.load = MagicMock(return_value=0)
        markov = Test
        fname = 'test'
        args = Namespace(progress=False, settings={})

        self.assertEqual(load(markov, fname, args), 0)
        self.assertFalse(getsize.called)
        self.assertFalse(op.called)
        self.assertFalse(bz2op.called)
        markov.load.assert_called_with(fname, args.settings)

        class Test2(MarkovJsonMixin):
            pass
        Test2.load = MagicMock(return_value=0)
        markov = Test2

        self.assertEqual(load(markov, fname, args), 0)
        getsize.assert_called_with(fname)
        op.assert_called_with(fname, 'rt')
        self.assertFalse(bz2op.called)

        getsize.return_value = IJSON_MIN_SIZE + 1
        self.assertEqual(load(markov, fname, args), 0)
        op.assert_called_with(fname, 'rb')
        self.assertFalse(bz2op.called)

        fname = '.bz2'
        getsize.return_value = IJSON_MIN_COMPRESSED_SIZE - 1
        op.called = False
        self.assertEqual(load(markov, fname, args), 0)
        bz2op.assert_called_with(fname, 'rt')
        self.assertFalse(op.called)

        getsize.return_value = IJSON_MIN_COMPRESSED_SIZE + 1
        self.assertEqual(load(markov, fname, args), 0)
        bz2op.assert_called_with(fname, 'rb')
        self.assertFalse(op.called)

        self.assertEqual(stdout.getvalue(), '')

    @patch('builtins.open', new_callable=mock_open)
    @patch('bz2.open', new_callable=mock_open)
    @patch('os.path.getsize', return_value=IJSON_MIN_SIZE - 1)
    @patch('sys.stdout', new_callable=StringIO)
    def testProgress(self, stdout, getsize, bz2op, op): # pylint:disable=unused-argument
        class Test(MarkovJsonMixin):
            pass
        Test.load = MagicMock(return_value=0)
        markov = Test
        fname = 'test'
        args = Namespace(progress=True, settings={})

        self.assertEqual(load(markov, fname, args), 0)
        self.assertNotEqual(stdout.getvalue(), '')

        class Test2():
            pass
        Test2.load = MagicMock(return_value=0)
        markov = Test2
        stdout.seek(0)
        stdout.truncate(0)
        self.assertEqual(load(markov, fname, args), 0)
        self.assertEqual(stdout.getvalue(), '')


class TestSave(TestCase):
    @patch('builtins.open', new_callable=mock_open)
    @patch('bz2.open', new_callable=mock_open)
    @patch('sys.stdout', new_callable=StringIO)
    def testOpen(self, stdout, bz2op, op):
        markov = MagicMock()
        fname = 'test'
        args = Namespace(progress=False)

        save(markov, fname, args)
        self.assertFalse(op.called)
        self.assertFalse(bz2op.called)
        markov.save.assert_called_with()

        markov = MagicMock(spec=MarkovJsonMixin)
        save(markov, fname, args)
        op.assert_called_with(fname, 'wt')
        self.assertFalse(bz2op.called)
        self.assertTrue(markov.save.called)

        markov.save.reset_mock()
        op.called = False
        save(markov, None, args)
        self.assertFalse(op.called)
        self.assertFalse(bz2op.called)
        markov.save.assert_called_with(stdout)

        fname = '.bz2'
        op.called = False
        markov.save.reset_mock()
        save(markov, fname, args)
        bz2op.assert_called_with(fname, 'wt')
        self.assertFalse(op.called)
        self.assertTrue(markov.save.called)

        self.assertEqual(stdout.getvalue(), '')

    @patch('builtins.open', new_callable=mock_open)
    @patch('bz2.open', new_callable=mock_open)
    @patch('sys.stdout', new_callable=StringIO)
    def testProgress(self, stdout, bz2op, op): # pylint:disable=unused-argument
        markov = MagicMock(spec=MarkovJsonMixin)
        fname = 'test'
        args = Namespace(progress=True)

        save(markov, fname, args)
        self.assertNotEqual(stdout.getvalue(), '')

        stdout.seek(0)
        stdout.truncate(0)
        save(markov, None, args)
        self.assertEqual(stdout.getvalue(), '')

        markov = MagicMock()
        save(markov, fname, args)
        self.assertEqual(stdout.getvalue(), '')


class TestSetArgs(TestCase):
    def testErrors(self):
        args = Namespace(output=sys.stdout, progress=True)
        with self.assertRaises(ValueError):
            set_args(args, ())

    def testType(self):
        args = Namespace()
        set_args(args, ())
        self.assertEqual(args.type, JSON)
        self.assertTrue(issubclass(args.markov, MarkovBase))
        self.assertTrue(issubclass(args.markov, MarkovJsonMixin))

        args = Namespace()
        set_args(args, (Mock,))
        self.assertEqual(args.type, JSON)
        self.assertTrue(issubclass(args.markov, Mock))
        self.assertTrue(issubclass(args.markov, MarkovBase))
        self.assertTrue(issubclass(args.markov, MarkovJsonMixin))

        args = Namespace(output='.db')
        set_args(args, ())
        self.assertEqual(args.type, SQLITE)
        self.assertTrue(issubclass(args.markov, MarkovBase))
        self.assertTrue(issubclass(args.markov, MarkovSqliteMixin))

        args = Namespace(output='.db', state='.json.bz2')
        set_args(args, ())
        self.assertEqual(args.type, JSON)

        args = Namespace(output='.db', state='.json.bz2', type='sqlite')
        set_args(args, ())
        self.assertEqual(args.type, SQLITE)

        args = Namespace(output='.db', state='.json.bz2', type='json')
        set_args(args, ())
        self.assertEqual(args.type, JSON)

    @patch('json.load', new=lambda x: x())
    def testSettings(self):
        args = Namespace()
        set_args(args, ())
        self.assertIsNone(args.settings)

        args = Namespace(settings=None)
        set_args(args, ())
        self.assertEqual(args.settings, {})

        s = list(range(3))
        mock = MagicMock(return_value=s)
        args = Namespace(settings=mock)
        set_args(args, ())
        self.assertEqual(args.settings, s)
        mock.close.assert_called_with()

class TestCheckOutputFormat(TestCase):
    def testError(self):
        tests = [
            ('test', -1),
            ('test', 0),
            ('test', 2),
            ('test%d%d', 2)
        ]
        for test in tests:
            with self.assertRaises(ValueError):
                check_output_format(*test)

    def testNoError(self): # pylint: disable=no-self-use
        tests = [
            ('test', 1),
            ('test%d', 2)
        ]
        for test in tests:
            check_output_format(*test)


class TestInFiles(TestCase):
    @patch('markovchain.cli.util.tqdm')
    def testNoProgress(self, tqdm):
        tests = [
            ([], True),
            (list(range(3)), False)
        ]
        for test in tests:
            with infiles(*test) as it:
                self.assertFalse(tqdm.called)
                self.assertIs(it, test[0])

    @patch('markovchain.cli.util.tqdm')
    def testProgress(self, tqdm):
        test = list(range(3))
        with infiles(test, True) as pbar:
            self.assertTrue(tqdm.called)
            self.assertTrue((test,) in tqdm.call_args)
            self.assertIsInstance(pbar, Mock)
        pbar.close.assert_called_with()


class TestOutFiles(TestCase):
    def testError(self):
        tests = [
            ('', -1, False),
            ('', 0, False)
        ]
        for test in tests:
            with self.assertRaises(ValueError), outfiles(*test):
                pass

    @patch('markovchain.cli.util.tqdm')
    def testNoProgress(self, tqdm):
        tests = [
            (('', 1, False), ['']),
            (('%d', 2, False), ['0', '1'])
        ]
        for test, res in tests:
            with outfiles(*test) as it:
                it = list(it)
                self.assertFalse(tqdm.called)
                self.assertEqual(it, res)

    @patch('markovchain.cli.util.tqdm')
    def testProgress(self, tqdm):
        test = ('%d', 3, True)
        res = ['0', '1', '2']
        with outfiles(*test) as pbar:
            self.assertTrue(tqdm.called)
            self.assertEqual(list(tqdm.call_args[0][0]), res)
            self.assertIsInstance(pbar, Mock)
        pbar.close.assert_called_with()



from math import floor
from itertools import islice
from functools import reduce
from PIL import Image

from ..scanner import Scanner
from ..util import fill, to_list, load
from .util import convert, palette as default_palette
from .traversal import Traversal, HLines

class ImageScanner(Scanner):
    """Image scanner class.

    Attributes
    ----------
    palette_image : `PIL.Image` or `None`
        Palette source image.
    resize : (`int`, `int`) or `None`
        If not None, resize images before scanning.
    min_size : `int`
        Minimum image size.
    convert : `int`
        Image conversion type.
    dither : `bool`
        If True, enable image dithering.
    """
    def __init__(self,
                 resize=None,
                 convert_type=1,
                 dither=False,
                 palette=None,
                 levels=1,
                 level_scale=4,
                 scale=Image.BICUBIC,
                 traversal=None):
        """Image scanner constructor.

        Parameters
        ----------
        resize : (`int`, `int`) or `None`, optional
            If not `None`, resize images before scanning (default: `None`).
        convert_type : `int`, optional
            Image conversion type (default: 1).
        dither : `bool`
            If `True`, enable image dithering (default: `False`).
        palette : `list` of `int`, optional
            Image palette (default: `markovchain.image.util.palette(8, 4, 8)`).
        levels : `int`, optional
            Number of scanner levels (default: 1).
        level_scale : `int` or `list` of `int`
            Level scale factors.
        scale : `int`, optional
            Scale quality (default: `PIL.Image.BICUBIC`).
        traversal : `dict` or `markovchain.image.traversal.Traversal` or `list` of (`dict` or `markovchain.image.traversal.Traversal`), optional
            Level traversals (default: `markovchain.image.traversal.HLines()`).
        """
        super().__init__()
        self._levels = None
        self._level_scale = None
        self._level_scale_max = None
        self._palette = None
        self._traversal = None
        self._traversal_max = None
        self.min_size = None
        self.palette_image = None
        self.resize = tuple(resize) if resize is not None else None
        self.palette = palette
        self.convert = convert_type
        self.levels = levels
        self.level_scale = level_scale
        self.dither = dither
        if isinstance(scale, str):
            scale = Image.__getattribute__(scale)
        self.scale = scale
        self.traversal = traversal

    @property
    def traversal(self):
        """`markovchain.image.traversal.Traversal`: Image traversal.
        """
        return self._traversal

    @traversal.setter
    def traversal(self, ts):
        ts = to_list(ts)

        for i, tr in enumerate(ts):
            ts[i] = load(tr, Traversal, HLines)

        if self.levels is not None:
            filled = fill(ts, self.levels, True)
        else:
            filled = ts

        self._traversal_max = ts if len(ts) > len(filled) else filled
        self._traversal = filled

    @property
    def levels(self):
        """`int`: Number of levels.
        """
        return self._levels

    @levels.setter
    def levels(self, levels):
        if levels <= 0:
            raise ValueError('level count <= 0')
        self._levels = levels
        if self._traversal_max is not None:
            self.traversal = self._traversal_max
        if self._level_scale_max is not None:
            self.level_scale = self._level_scale_max

    @property
    def level_scale(self):
        """`list` of `int`: Level scale factors.
        """
        return self._level_scale

    @level_scale.setter
    def level_scale(self, scale):
        scale = to_list(scale)

        if any(x <= 1 for x in scale):
            raise ValueError('level scale <= 1: {0}'.format(scale))

        if self.levels is not None:
            filled = fill(scale, self.levels - 1)
            if filled:
                size = reduce(lambda x, y: x * y, filled)
            else:
                size = 1
        else:
            filled = scale
            size = 1

        self._level_scale_max = scale if len(scale) > len(filled) else filled
        self._level_scale = filled
        self.min_size = size

    @property
    def palette(self):
        """`list` of `int` : Image palette.
        """
        return self._palette

    @palette.setter
    def palette(self, palette):
        if palette is None:
            palette = [8, 4, 8]
        if len(palette) == 3:
            palette = default_palette(*palette)
        self.palette_image = Image.new('P', (1, 1))
        self.palette_image.putpalette(palette)
        self._palette = palette

    def input(self, img):
        """Resize input image if necessary.

        Parameters
        ----------
        img : `PIL.Image`
            Input image.

        Raises
        ------
        ValueError
            If input image is too small.

        Returns
        -------
        `PIL.Image`
            Resized image or input image.
        """
        img_width, img_height = img.size

        if self.resize:
            width, height = self.resize
            scale = min(width / img_width, height / img_height)
            img = img.resize((floor(img_width * scale),
                              floor(img_height * scale)))
            img_width, img_height = img.size

        if img_width < self.min_size or img_height < self.min_size:
            raise ValueError('input image is too small: {0} x {1} < {2} x {2}'
                             .format(img_width, img_height, self.min_size))

        return img

    def set_palette(self, img):
        """Set image palette.

        Parameters
        ----------
        img : `PIL.Image`
            Input image.

        Returns
        -------
        `PIL.Image`
            Converted image.
        """
        return convert(self.convert, img, self.palette_image, self.dither)

    def level(self, img, level):
        """Get image level.

        Parameters
        ----------
        img : `PIL.Image`
            Input image.
        level : `int`
            Level number.

        Returns
        -------
        `PIL.Image`
            Converted image.
        """
        if level < self.levels - 1:
            width, height = img.size
            scale = reduce(lambda x, y: x * y,
                           islice(self.level_scale, level, self.levels))
            img = img.resize((width // scale, height // scale), self.scale)

        return self.set_palette(img)

    def _scan_level(self, level, prev, img):
        """Scan a level.

        Parameters
        ----------
        level : `int`
            Level number.
        prev : `PIL.Image` or None
            Previous level image or None if level == 0.
        img : `PIL.Image`
            Current level image.

        Returns
        -------
        `generator` of (`str` or `markovchain.scanner.Scanner.END` or (`markovchain.scanner.Scanner.START`, `str`))
            Token generator.
        """
        if level == 0:
            width, height = img.size
        else:
            width, height = prev.size

        tr = self.traversal[0](width, height, ends=(level == 0))
        if level == 0:
            for xy in tr:
                if xy is None:
                    yield self.END
                else:
                    yield '%02X' % img.getpixel(xy)
            yield self.END
        else:
            scale = self.level_scale[level - 1]
            for xy in tr:
                x0 = xy[0] * scale
                y0 = xy[1] * scale
                start = (
                    self.START,
                    '%02X%02X' % (level - 1, prev.getpixel(xy))
                )
                yield start
                for dxy in self.traversal[level](scale, scale, True):
                    if dxy is None:
                        yield start
                    yield '%02X%02X' % (
                        level, img.getpixel((x0 + dxy[0], y0 + dxy[1]))
                    )
                yield self.END

    def __call__(self, img, part=False):
        """Scan an image.

        Parameters
        ----------
        img : `PIL.Image`
            Image to scan.
        part : `bool`, optional
            True if data is partial.

        Raises
        ------
        NotImplementedError
            If `part` is `True`.

        Returns
        -------
        `generator` of (`str` or `markovchain.scanner.Scanner.END` or (`markovchain.scanner.Scanner.START`, `str`))
            Token generator.
        """
        if part:
            raise NotImplementedError()

        prev = None
        resized = self.input(img)

        for level in range(self.levels):
            img = self.level(resized, level)
            yield self._scan_level(level, prev, img)
            prev = img

    def __eq__(self, scanner):
        return (self.resize == scanner.resize
                and self.convert == scanner.convert
                and self.dither == scanner.dither
                and self.palette == scanner.palette
                and self.traversal == scanner.traversal
                and self.levels == scanner.levels
                and self.level_scale == scanner.level_scale
                and self.scale == scanner.scale)

    def save(self):
        """Convert to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().save()
        data['resize'] = list(self.resize) if self.resize is not None else None
        data['convert_type'] = self.convert
        data['dither'] = self.dither
        data['palette'] = self.palette
        data['traversal'] = [t.save() for t in self.traversal]
        data['levels'] = self.levels
        data['level_scale'] = self.level_scale
        data['scale'] = self.scale
        return data



#!/usr/bin/env python3

import os
from unittest import TestLoader
from setuptools import setup, find_packages

def tests():
    return TestLoader().discover('tests')

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(BASE_DIR, 'README.rst')) as fp:
        README = fp.read()
except FileNotFoundError:
    README = ''

setup(name='markovchain',
      version='0.1.0',
      description='Markov chain generator',
      long_description=README,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: End Users/Desktop',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Multimedia :: Graphics',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Text Processing',
          'Topic :: Utilities'
      ],
      keywords='markov chain generator',
      url='https://github.com/dead-beef/markovchain',
      author='dead-beef',
      author_email='contact@dead-beef.tk',
      license='MIT',
      packages=find_packages(include=('markovchain*',)),
      entry_points={
          'console_scripts': ['markovchain=markovchain.cli:main'],
      },
      test_suite='setup.tests',
      install_requires=['tqdm', 'ijson'],
      extras_require={
          'image': ['pillow'],
          'dev': [
              'pillow',
              'coverage',
              'sphinx',
              'sphinx_rtd_theme'
          ]
      },
      python_requires='>=3',
      include_package_data=True,
      zip_safe=False)



import json
import sqlite3
from random import randint
from itertools import chain, islice

from .util import extend

class MarkovSqliteMixin:
    """Markov chain SQLite data mixin.

    Attributes
    ----------
    db : `sqlite3.Connection`
        Database connection.
    cursor
        Database cursor.
    """
    def __init__(self, db=':memory:', *args, **kwargs):
        """Markov chain SQLite data constructor.

        Parameters
        ----------
        db : `str` or `sqlite3.Connection`, optional
            Database path or connection (default: ':memory:').
        """
        super().__init__(*args, **kwargs)
        if isinstance(db, str):
            db = sqlite3.connect(db, isolation_level='IMMEDIATE')
        self.db = db
        self.cursor = db.cursor()
        self.create_node_tables()

    def __eq__(self, markov):
        return super().__eq__(markov)

    def replace_state_separator(self, old_separator, new_separator):
        """Replace state separator.

        Parameters
        ----------
        old_separator : `str`
            Old state separator.
        new_separator : `str`
            New state separator.
        """
        self.cursor.execute(
            'UPDATE nodes SET value = replace(value, ?, ?)',
            (old_separator, new_separator)
        )

    def links(self, links):
        """Update links.

        Parameters
        ----------
        links : `generator` of (`islice` of `str`, `str`)
            Links to add.
        """
        for src, dst in links:
            src = list(src)
            source = self.get_node(self.separator.join(src))
            target = self.get_node(self.separator.join(
                chain(islice(src, 1, None), (dst,))
            ))
            self.cursor.execute(
                'UPDATE links SET count = count + 1 WHERE source=? AND target=?',
                (source, target)
            )
            self.cursor.execute(
                '''INSERT INTO links (source, target, value)
                   SELECT ?, ?, ?
                   WHERE (SELECT Changes() = 0)''',
                (source, target, dst)
            )

    def random_link(self, state):
        """Get a random link.

        Parameters
        ----------
        state : `int` or `deque` of `str`
            Link source ID or name.

        Returns
        -------
        (`str`, `int`)
            Link value and next source ID.
        """
        if not isinstance(state, int):
            self.cursor.execute(
                'SELECT id FROM nodes WHERE value=?',
                (self.separator.join(state),)
            )
            x = state
            state = self.cursor.fetchone()
            if state is None:
                return None, None
            state = state[0]

        self.cursor.execute(
            'SELECT count, value, target FROM links WHERE source=?',
            (state,)
        )
        links = self.cursor.fetchall()
        if not links:
            return None, None
        count = sum(link[0] for link in links)
        x = randint(0, count - 1)
        for count, value, target in links:
            if x < count:
                return value, target
            x -= count
        raise RuntimeError('no link')

    def get_tables(self):
        """Get all table names.

        Returns
        -------
        `set` of `str`
        """
        self.cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table"'
        )
        return set(x[0] for x in self.cursor.fetchall())

    def get_node(self, value):
        """Get node ID by value.

        If a node with the specified value does not exist,
        create it and return its ID.

        Parameters
        ----------
        value : `str`
            Node value.

        Returns
        -------
        `int`
            Node ID.
        """
        while True:
            self.cursor.execute(
                'SELECT id FROM nodes WHERE value=?',
                (value,)
            )
            node = self.cursor.fetchone()
            if node is not None:
                return node[0]
            self.cursor.execute(
                'INSERT INTO nodes (value) VALUES (?)',
                (value,)
            )

    def update_main_table(self):
        """Write generator settings to database.
        """
        data = (json.dumps(self.get_save_data()),)
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS main (
                settings TEXT NOT NULL DEFAULT "{}"
            )
        ''')
        self.cursor.execute('SELECT * FROM main')
        if self.cursor.fetchall() == []:
            self.cursor.execute('INSERT INTO main (settings) VALUES (?)', data)
        else:
            self.cursor.execute('UPDATE main SET settings=?', data)

    def create_node_tables(self):
        """Create node and link tables if they don't exist.
        """
        self.cursor.execute('PRAGMA foreign_keys=1')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS nodes (
                id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
                value TEXT NOT NULL
            )
        ''')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS links (
                source REFERENCES nodes (id),
                target REFERENCES nodes (id),
                value TEXT NOT NULL,
                count INTEGER NOT NULL DEFAULT 1
            )
        ''')
        self.cursor.execute(
            'CREATE UNIQUE INDEX IF NOT EXISTS node ON nodes (value)'
        )
        self.cursor.execute(
            'CREATE INDEX IF NOT EXISTS link_source ON links (source)'
        )
        #self.cursor.execute(
        #    'CREATE INDEX IF NOT EXISTS link_target ON links (target)'
        #)

    @classmethod
    def load(cls, fp, override=None):
        """Load a generator.

        Parameters
        ----------
        fp : `file` or `str`
            Input file or file path.
        override : `dict` or `None`, optional
            Changes to loaded data (default: `None`).

        Returns
        -------
        Loaded generator.
        """
        if not isinstance(fp, str):
            fp.close()
            fp = fp.name

        db = sqlite3.connect(fp)

        try:
            cursor = db.cursor()
            cursor.execute('SELECT settings FROM main')
            args = json.loads(cursor.fetchone()[0])
        except sqlite3.OperationalError:
            args = {}
            raise

        if override is not None:
            extend(args, override)

        args['db'] = db

        return cls(**args)

    def save(self):
        """Save the generator.
        """
        self.update_main_table()
        self.db.commit()



from itertools import chain, repeat, islice
from PIL import Image

from .scanner import ImageScanner
from .util import palette as default_palette
from ..parser import LevelParser
from ..util import fill


class MarkovImageMixin:
    """Markov chain image generator mixin.
    """

    DEFAULT_SCANNER = ImageScanner
    DEFAULT_PARSER = LevelParser

    def __init__(self,
                 levels=1,
                 palette=None,
                 *args, **kwargs):
        """Markov chain image generator constructor.

        Parameters
        ----------
        levels : `int`, optional
            Number of levels.
        palette : `list` of `int`, optional
            Image palette (default: `markovchain.image.util.palette(8, 4, 8)`).
        """
        super().__init__(*args, **kwargs)
        if palette is None:
            try:
                palette = self.scanner.palette
            except AttributeError:
                palette = default_palette(8, 4, 8)
        elif len(palette) == 3:
            palette = default_palette(*palette)
        self.palette = palette
        self.levels = levels

    @property
    def levels(self):
        """`int` : Number of levels.
        """
        return self._levels

    @levels.setter
    def levels(self, levels):
        if self.scanner is not None:
            self.scanner.levels = levels
        if self.parser is not None:
            self.parser.levels = levels
        self._levels = levels

    def _imgdata(self, width, height, state_size=None, start=None):
        """Generate image pixels.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        state_size : `int` or `None`, optional
            State size to use for generation (default: `None`).
        start : `str` or `None`, optional
            Initial state (default: `None`).

        Raises
        ------
        RuntimeError
            If generator is empty.

        Returns
        -------
        `generator` of `int`
            Pixel generator.
        """
        maxlength = width * height
        if maxlength > 0 and start is not None:
            yield int(start[-2:], 16)
            maxlength -= 1
        while maxlength > 0:
            prevmaxlength = maxlength
            for pixel in self.generate(maxlength=maxlength,
                                       state_size=state_size,
                                       start=start):
                yield int(pixel[-2:], 16)
                maxlength -= 1
            if prevmaxlength == maxlength:
                if start is not None:
                    pixel = int(start[-2:], 16)
                    yield from repeat(pixel, maxlength)
                else:
                    raise RuntimeError('empty generator')

    def image(self, width, height,
              state_size=None,
              start=None, levels=None,
              start_level=-1, start_image=None):
        """Generate an image.

        Parameters
        ----------
        width : `int`
            Image width.
        height : `int`
            Image height.
        state_size : `int` or `None`, optional
            State size to use for generation (default: `None`).
        start : `str` or `None`, optional
            Initial state (default: `None`).
        levels : `int`, optional
            Number of generated levels (default: `self.scanner.levels`).
        start_level : `int`, optional
            Initial level (default: -1).
        start_image : `PIL.Image` or `None`
            Initial level image (default: `None`).

        Returns
        -------
        `PIL.Image`
            Generated image.
        """
        if levels is None:
            levels = self.scanner.levels
        elif levels <= 0 or levels > self.levels:
            raise ValueError('invalid level count: {0}'.format(levels))

        if start_image is not None:
            start_image = self.scanner.set_palette(start_image)
        else:
            start_level = -1

        if start_level is None or start_level < 0:
            start_level = -1
            start_image = None
        elif start_level + 1 >= levels:
            return start_image

        state_sizes = fill(state_size, levels)
        #if any(sz is not None and sz not in parser.state_sizes
        #       for sz, parser in zip(state_sizes, self.parser.parsers)):
        #    raise ValueError('invalid state sizes: {0}: not in {1}'.format(
        #        state_sizes,
        #        [parser.state_sizes for parser in self.parser.parsers]
        #    ))

        prev = start_image
        gen_levels = islice(enumerate(state_sizes), start_level + 1, levels)

        if prev is not None:
            width, height = prev.size

        for level, state_size in gen_levels:
            if state_size is None:
                try:
                    state_size = self.parser.parsers[level].state_sizes[0]
                except AttributeError:
                    pass

            if prev is not None:
                scale = self.scanner.level_scale[level - 1]
                width *= scale
                height *= scale

            ret = Image.new('P', (width, height))
            ret.putpalette(self.palette)

            if prev is None:
                tr = self.scanner.traversal[0](width, height, ends=False)
                data = self._imgdata(width, height, state_size, start)
                for xy, pixel in zip(tr, data):
                    ret.putpixel(xy, pixel)
            else:
                tr = self.scanner.traversal[0](
                    prev.size[0],
                    prev.size[1],
                    ends=False
                )
                for xy in tr:
                    start = '%02X%02X' % (level - 1, prev.getpixel(xy))
                    x0, y0 = xy
                    x0 *= scale
                    y0 *= scale
                    data = self._imgdata(scale, scale, state_size, start)
                    blk = self.scanner.traversal[level](scale, scale, False)
                    for pixel, (x1, y1) in zip(data, blk):
                        ret.putpixel((x0 + x1, y0 + y1), pixel)

            prev = ret

        return ret

    def data(self, data, part=False):
        """Parse data and update links.

        Parameters
        ----------
        data : `PIL.Image`
            Data to parse.
        part : `bool`, optional
            `True` if data is partial (default: `False`).
        """
        if isinstance(self.parser, LevelParser):
            self.links(self.parser(self.scanner(data, part), part))
        else:
            data = chain.from_iterable(self.scanner(data, part))
            self.links(self.parser(data, part))

    def get_save_data(self):
        """Convert generator settings to JSON.

        Returns
        -------
        `dict`
            JSON data.
        """
        data = super().get_save_data()
        data['palette'] = self.palette
        data['levels'] = self.levels
        return data

    def __eq__(self, markov):
        return (self.palette == markov.palette
                and self.levels == markov.levels)



from unittest import TestCase
from PIL import Image

from markovchain import MarkovBase, MarkovJsonMixin, Scanner, Parser
from markovchain.image import MarkovImageMixin, HLines, VLines


class TestMarkovImage(TestCase):
    class Markov(MarkovImageMixin, MarkovJsonMixin, MarkovBase):
        pass

    @classmethod
    def setUpClass(cls):
        palette = [
            0x00, 0x00, 0x00,
            0x44, 0x44, 0x44,
            0xaa, 0xaa, 0xaa,
            0xdd, 0xdd, 0xdd
        ]
        palette.extend(0 for _ in range((256 - len(palette)) * 3))
        cls.palette = palette

    def test_properties(self):
        m = self.Markov()
        self.assertEqual(m.scanner.levels, m.levels)
        self.assertEqual(m.parser.levels, m.levels)
        m.levels = 3
        self.assertEqual(m.scanner.levels, m.levels)
        self.assertEqual(m.parser.levels, m.levels)
        with self.assertRaises(ValueError):
            m.levels = -1
        self.assertEqual(m.scanner.levels, m.levels)
        self.assertEqual(m.parser.levels, m.levels)
        self.assertEqual(m.levels, 3)

    def test_generate(self):
        scanner = Scanner(lambda x: x)
        scanner.traversal = [HLines()]
        scanner.levels = 1
        scanner.level_scale = []

        m = self.Markov(palette=self.palette, scanner=scanner, parser=Parser())

        with self.assertRaises(RuntimeError):
            m.image(2, 2)

        m.data([['00', '01', '02', '03']])

        data = list(m.image(2, 2).getdata())
        self.assertEqual(data, [0, 1, 2, 3])

        data = list(m.image(4, 2).getdata())
        self.assertEqual(data, [0, 1, 2, 3, 0, 1, 2, 3])

    def test_generate_levels(self):
        scanner = Scanner(lambda x: x)
        scanner.traversal = [HLines(), VLines()]
        scanner.levels = 2
        scanner.level_scale = [2]
        scanner.set_palette = lambda img: img

        m = self.Markov(levels=2, palette=self.palette, scanner=scanner)

        with self.assertRaises(RuntimeError):
            m.image(2, 2)

        m.data([
            ['00', '01', '02', '03',],
            [(Scanner.START, '0000'), '0101',
             (Scanner.START, '0001'), '0102',
             (Scanner.START, '0002'), '0103',
             (Scanner.START, '0003'), '0100',]
        ])

        data = list(m.image(2, 2, levels=1).getdata()) # pylint: disable=no-member
        self.assertEqual(data, [0, 1, 2, 3])

        data = list(m.image(2, 2, levels=2).getdata()) # pylint: disable=no-member
        self.assertEqual(
            data,
            [
                0, 1, 1, 2,
                1, 1, 2, 2,
                2, 3, 3, 0,
                3, 3, 0, 0
            ]
        )

        with self.assertRaises(ValueError):
            m.image(2, 2, levels=3)
        with self.assertRaises(ValueError):
            m.image(2, 2, levels=0)

        img = Image.new('P', (1, 1))
        img.putpalette(self.palette)
        img.putpixel((0, 0), 1)

        data = list(m.image(
            2, 1, levels=1,
            start_level=-5, start_image=img
        ).getdata()) # pylint: disable=no-member
        self.assertEqual(data, [0, 1])

        data = list(m.image(
            None, None, levels=1,
            start_level=0, start_image=img
        ).getdata()) # pylint: disable=no-member
        self.assertEqual(data, [1])

        data = list(m.image(
            None, None, levels=1,
            start_level=2, start_image=img
        ).getdata()) # pylint: disable=no-member
        self.assertEqual(data, [1])

        data = list(m.image(
            None, None, levels=2,
            start_level=0, start_image=img
        ).getdata()) # pylint: disable=no-member
        self.assertEqual(data, [1, 2, 2, 2])

    def test_save_load(self):
        m = self.Markov()
        saved = m.get_save_data()
        loaded = self.Markov(**saved)
        self.assertEqual(m, loaded)
